from dataclasses import dataclass, field
import numpy as np
from typing import Tuple

import forsys.virtual_edges as ve

@dataclass
class SmallEdge:
    """
    Class representation for small edges. This are defined as the lines joining two vertices.

    :param id: Unique identifier of the small edge
    :type id: int
    :param v1: First vertex of the edge
    :type v1: object
    :param v2: Second vertex of the edge
    :type v2: object
    :param tension: Inferred tension in this edge. Defaults to 0
    :type tension: float
    :param gt: Ground truth tension of this edge. Defaults to 0
    :type gt: float
    """
    id: int
    v1: object
    v2: object
    tension: float = 0
    gt: float = 0

    def __post_init__(self):
        """ Constructor method
        """
        for v in self.get_vertices_array():
            v.add_edge(self.id)
        assert self.v1.id != self.v2.id, f"edge {self.id} with the same vertex twice"

    def __del__(self):
        for v in self.get_vertices_array():
            if self.id in v.ownEdges:
                v.remove_edge(self.id)

    def get_vertices_id(self) -> list:
        """
        Get IDs of the vertices that form the edge

        :return: Array with the two IDs of the vertices
        :rtype: list
        """
        return [self.v1.id, self.v2.id]

    def get_other_vertex_id(self, one: int) -> int:
        """
        Get the vertex ID of the other vertex in the edge, given one of them

        :param one: ID of the cell to be added
        :type one: int
        :return: ID of the vertex
        :rtype: int
        """
        assert one in self.get_vertices_id(), "vertex not in asked edge"
        return self.v1.id if one == self.v2.id else self.v2.id

    def get_vertices_array(self) -> list:
        """
        Get the IDs of the vertices that form the edge

        :return: [id_1, id_2] array for both vertices' IDs
        :rtype: list
        """
        return [self.v1, self.v2]

    def replace_vertex(self, vold: object, vnew: object) -> None:
        """
        Replace one of the vertices in the edge with a new one

        :param vold: Vertex object to be replaced
        :type vold: object
        :param vnew: Vertex object to be added
        :type vnew: object
        """
        who = 0 if self.v1.id == vold.id else 1
        self.get_vertices_array()[who].remove_edge(self.id)
        if who == 0:
            self.v1 = vnew
        else:
            self.v2 = vnew
        self.get_vertices_array()[who].add_edge(self.id)

    def get_vector(self) -> list:
        """
        Get the vector connecting the two vertices of the edge.

        :return: Vector in the (0, 0) frame of reference
        :rtype: list
        """
        vector = [self.v2.x - self.v1.x, self.v2.y - self.v1.y]
        return vector


@dataclass
class BigEdge:
    """
    Class representation for big edges. This are defined as a set of SmallEdges.
    Could be only one small edge in special cases. This allows for curved representations
    of the cellular membrane.

    :param big_edge_id: Unique identifier of the big edge
    :type big_edge_id: int
    :param vertices: List of vertices form the big edge
    :type vertices: list
    :param tension: Inferred tension in this edge. Defaults to 0
    :type tension: float
    :param gt: Ground truth tension of this edge. Defaults to 0
    :type gt: float
    :param external: True if this big edge is external to the tissue
    :type external: bool
    """
    big_edge_id: int
    vertices: list

    tension: float = 0.0
    gt: float = 0.0
    external: bool = False

    def __post_init__(self):
        """ Constructor method
        """
        for vertex in self.vertices:
            vertex.add_big_edge(self.big_edge_id)

        self.xs = [vertex.x for vertex in self.vertices]
        self.ys = [vertex.y for vertex in self.vertices]
        self.edges = [list(set(self.vertices[vid].ownEdges) &
                           set(self.vertices[vid + 1].ownEdges))[0]
                      for vid in range(0, len(self.vertices) - 1)]

        if len(self.vertices) == 2:
            self.own_cells = list(set(self.vertices[0].ownCells) & set(self.vertices[1].ownCells))
        else:
            self.own_cells = self.vertices[(len(self.vertices) - 1) // 2].ownCells
        vertices_own_cells = [len(vertex.ownCells) < 2 for vertex in self.vertices]
        has_junction_at_extremes = True if len(self.vertices[0].ownCells) > 2 \
                                           or len(self.vertices[-1].ownCells) > 2 else False
        self.external = True if np.any(vertices_own_cells) or not has_junction_at_extremes \
            else False
        

    def calculate_curvature(self) -> float:
        """
        Get the curvature of the big edge

        :return: Object reference for the next vertex in the ordering
        :rtype: object
        """
        dx_dt = np.gradient(self.xs)
        dy_dt = np.gradient(self.ys)
        d2x_dt2 = np.gradient(dx_dt)
        d2y_dt2 = np.gradient(dy_dt)
        curvature = (d2x_dt2 * dy_dt - dx_dt * d2y_dt2) / ((dx_dt**2 + dy_dt**2)**1.5)

        return curvature

    def calculate_total_curvature(self, normalized=True) -> float:
        """
        Get the total curvature of the big edge normalized by its length

        :param normalized: Current vertex after which the next is calculated
        :type normalized: bool
        :return: Object reference for the next vertex in the ordering
        :rtype: object
        """
        curvatures = self.calculate_curvature()
        ds = np.sqrt(np.diff(self.xs)**2 + np.diff(self.ys)**2)
        total_curvature = np.sum((curvatures[1:] + curvatures[:-1])/2 * ds)
        if normalized:
            curve_length = np.sum(np.sqrt(np.diff(self.xs)**2 + np.diff(self.ys)**2))
            total_curvature = total_curvature / curve_length
        return total_curvature


    def get_versor_from_vertex(self, vid: int, method: str = "edge", cell: object = None, fit_method:str = "dlite") -> list:
        """
        Get the versor corresponding to the vectorial representation of the big edge.
        
        :param vid: The ID of the vertex.
        :type vid: int
        :param method: The method used to calculate the vector representation of the big edge. Default is "edge".
        :type method: str, optional
        :param cell: The cell object. Default is None.
        :type cell: object, optional
        :param fit_method: The fit method used. Default is "dlite".
        :type fit_method: str, optional
        :return: The versor corresponding to the vectorial representation of the big edge.
        :rtype: list
        """
        vector = self.get_vector_from_vertex(vid, method, cell, fit_method)
        versor = vector / np.linalg.norm(vector)
        return versor

    def get_vector_from_vertex(self, vid: int, method: str = "edge", cell: object = None, fit_method: str="dlite") -> list:
        """
        Get the vector representation for the big edge

        :param vid: ID of the vertex from where the vector is defined
        :type vid: int
        :param method: Method to calculate representation. "edge" gives the circular fit to the edge, \
            "cell" and passing a cell object gives a circular fit to the given cell.
        :type method: str
        :param cell: Object reference of the cell to calculate the curvature if necessary
        :type cell: object
        :return: x, y coordinates of the vector
        :rtype: list
        """
        vobject = self.get_vertex_object_by_id(vid)
        if method == "edge":
            xc, yc = ve.calculate_circle_center(self.vertices, method=fit_method)
        elif method == "cell" and cell:
            xc, yc = cell.center_x, cell.center_y
        else:
            raise Exception("Method for versor doesn't exist or cell missing")

        vector = np.array((- (vobject.y - yc), (vobject.x - xc)))

        correct_sign = self.get_versor_sign(vid)
        if np.any(np.sign(vector) != correct_sign):
            correction = correct_sign * np.sign(vector)
            vector = vector * correction
        return vector

    def get_vertex_object_by_id(self, vid: int) -> object:
        """
        Get the vertex object corresponding to the given ID

        :param vid: Current vertex after which the next is calculated
        :type vid: int
        :return: Object reference for required vertex
        :rtype: object
        """
        vertex_object = list(filter(lambda x: x.id == vid, self.vertices))
        assert len(vertex_object) == 1, "More than one or no vertex with the same ID"
        return vertex_object[0]

    def get_vertices_ids(self) -> list:
        """
        Get all IDs of the vertices that form the big edge

        :return: Array with all vertices' IDs
        :rtype: list
        """
        return [vertex.id for vertex in self.vertices]

    def get_versor_sign(self, vid: int) -> list:
        """
        Get the correction sign for each
        componenet of the vector

        :param vid: Vertex to calculate the sign from
        :type vid: int
        :return: Array with the sign corrections
        :rtype: list
        """
        versor = self.get_straight_edge_versor_from_vid(vid)
        signs = np.sign(versor)
        return [1.0 if sign == 0 else sign for sign in signs]


    def get_straight_edge_versor_from_vid(self, vid: int) -> list:
        all_vertices_ids = self.get_vertices_ids()
        v0 = self.get_vertex_object_by_id(vid)
        # connected to ?
        if all_vertices_ids[0] == vid:
            next_vid = all_vertices_ids[1]
        elif vid == all_vertices_ids[-1]:
            next_vid = all_vertices_ids[-2]
        else:
            raise ("Vertex ID is not in a junction")

        v1 = self.get_vertex_object_by_id(next_vid)
        return [v1.x - v0.x, v1.y - v0.y]

    def get_length(self):
        """
        Get the length of the big edge

        :return: Length of the big edge
        :rtype: float
        """
        return np.sum(np.sqrt(np.diff(self.xs)**2 + np.diff(self.ys)**2))



import matplotlib.pyplot as plt
import matplotlib as mpl
import os
import numpy as np
import forsys as fs

import forsys.virtual_edges as ve
import forsys.frames as fframes
import forsys.edge as fedges
import forsys.time_series as ftimes
import forsys.myosin as fmyo
from typing import Union, Tuple
from copy import deepcopy
import PIL
import tifffile as tif


def plot_with_stress_custom(frame: fframes.Frame, step: str, folder: str, stress_dictionary: dict, 
                            earr: list, versors: Union[bool, list]=False, maxForce: float=None, 
                            minForce: float=None, normalized: Union[bool, str]=False, mirror_y: bool=False, **kwargs) -> None:
    """Generate a plot on a given mesh using a custom force dictionary.

   :param frame: Frame to plot
    :type frame: fframes.Frame
    :param step: Step at which to create the plot. It is used as the file name
    :type step: int
    :param folder: Folder to save the plot to
    :type folder: str
    :param stress_dictionary: Custom dictionary of stresses
    :type stress_dictionary: dict
    :param earr: List of big edges in the system
    :type earr: list
    :param versors: Versors to plot if necessary. Useful when external stresses are used, defaults to False
    :type versors: Union[bool, list], optional
    :param maxForce: Max force to use in the normalization, defaults to None
    :type maxForce: float, optional
    :param minForce: Min force in the normalization, defaults to None
    :type minForce: float, optional
    :param normalized: "max" normalizes by the maximum force; "normal" normalizes to a \
        normal distribution centered in zero, defaults to False
    :type normalized: Union[bool, str], optional
    :param mirror_y: Plot mirroring the y component, defaults to False
    :type mirror_y: bool, optional
    """
    plt.close()
    if not os.path.exists(folder):
        os.makedirs(folder)
    jet = plt.get_cmap('jet')

    _, ax = plt.subplots(1,1)

    if not os.path.exists(folder):
        os.makedirs(folder)

    if maxForce == None:
        maxForce = max(stress_dictionary.values())
    if minForce == None:
        minForce = min(stress_dictionary.values())
    
    allValues = np.array(list(stress_dictionary.values())).flatten()
    aveForce = np.mean(allValues)
    stdForce = np.std(allValues)

    already_plotted = []

    for name, force in stress_dictionary.items():
        if not isinstance(name, str):
            try:
                current = earr[name]
            except IndexError:
                print(f"IndexError: trying big edge {name} in an earr of length {len(earr)}")
                pass
            for j in range(0, len(current)):
                eBelong = frame.vertices[current[j]].ownEdges
                for e in eBelong:
                    edge = frame.edges[e]
                    try:
                        if len(set(edge.get_vertices_id()) & set([current[j], current[j+1]])) == 2:
                            if normalized == "normal":
                                forceValToPlot = (force - aveForce) / stdForce
                            elif normalized == "max":
                                forceValToPlot = force / maxForce
                            else:
                                forceValToPlot = force/maxForce

                            color = jet(forceValToPlot)
                            plt.plot(   (edge.v1.x, edge.v2.x),
                                        (edge.v1.y, edge.v2.y),
                                        color=color, linewidth=3)
                            already_plotted.append(edge.id)
                    except IndexError:
                        pass
                    
    for edge in frame.edges.values():
        if edge.id not in already_plotted:
            plt.plot(   (edge.v1.x, edge.v2.x),
                        (edge.v1.y, edge.v2.y),
                        color="black", linewidth=0.5, alpha=0.6)

    plt.axis('off')
    if mirror_y:
        plt.gca().invert_yaxis()
    if kwargs.get("colorbar", False):
        sm = plt.cm.ScalarMappable(cmap="jet", norm=plt.Normalize(vmin=0, vmax=1))
        plt.colorbar(sm)

    name = os.path.join(folder, str(step) + ".png")

    plt.tight_layout()
    plt.savefig(name, dpi=500)

    plt.close()


def plot_inference(frame: fframes.Frame, pressure: bool = False,
                   ground_truth: bool = False, maxForce: float = None,
                   minForce: float = None,
                   normalized: Union[bool, str] = False,
                   ax: mpl.axes.Axes = None,
                   **kwargs) -> Tuple:
    """Generate a plot of a tissue using the inferred tensions and pressures and return the plot object

    :param frame: Frame to plot
    :type frame: fframes.Frame
    :param pressure: If True pressures are plotted in the inner space of the cells.
    :type pressure: bool
    :param ground_truth: False plots inferred stresses, True will plot the ground truth value assigned to each edge
    :param maxForce: Max force to use in the normalization, defaults to None
    :type maxForce: float, optional
    :param minForce: Min force in the normalization, defaults to None
    :type minForce: float, optional
    :param normalized: "max" normalizes by the maximum force; "normal" normalizes to a \
        normal distribution centered in zero, defaults to False
    :type normalized: Union[bool, str], optional
    :return: Matplotlib plot object
    :rtype: Tuple
    """
    if not ax:
        _, ax = plt.subplots(1, 1)
    if ground_truth:
        perturbation = 0.0
        all_forces = [edge.gt + np.random.uniform(-edge.gt*perturbation, edge.gt*perturbation) for edge in frame.edges.values() if edge.gt != 0]
    else:
        all_forces = [edge.tension for edge in frame.edges.values() if edge.tension != 0]

    if maxForce is None:
        maxForce = max(all_forces)
    if minForce is None:
        minForce = min(all_forces)

    aveForce = np.mean(all_forces)
    stdForce = np.std(all_forces)

    for _, edge in frame.edges.items():
        if edge.tension != 0:
            observable = edge.tension if not ground_truth else edge.gt
        else:
            observable = edge.tension if not ground_truth else edge.gt
        if observable <= 0:
            ax.plot((edge.v1.x, edge.v2.x),
                    (edge.v1.y, edge.v2.y),
                    color="black", linewidth=0.5, alpha=0.6)
        else:
            if normalized == "normal":
                forceValToPlot = (observable - aveForce) / stdForce
            elif normalized == "max":
                forceValToPlot = observable / maxForce
                min_stress = 0
                max_stress = 1
            elif normalized == "relative":
                forceValToPlot = (observable + minForce) / maxForce
            elif normalized == "absolute":
                forceValToPlot = observable / kwargs.get("max_stress", 2)
                min_stress = 0
                max_stress = 2
            else:
                forceValToPlot = observable/maxForce
                min_stress = 0
                max_stress = 1

            cmap_stress_name = kwargs.get("cmap_stress", "magma")
            cmap_pressure_name = kwargs.get("cmap_pressure", "cividis")
            cmap_stress = plt.get_cmap(cmap_stress_name)
            # cmap_pressure = plt.get_cmap(cmap_pressure_name)

            color_stress = cmap_stress(forceValToPlot)
            stress_cb = plt.cm.ScalarMappable(cmap=cmap_stress_name,
                                              norm=plt.Normalize(vmin=min_stress,
                                                                 vmax=max_stress))
            ax.plot((edge.v1.x, edge.v2.x),
                    (edge.v1.y, edge.v2.y),
                    color=color_stress, linewidth=2)

    if pressure:
        pressures = frame.get_pressures()["pressure"]
        pressures_cb = plt.cm.ScalarMappable(cmap=cmap_pressure_name,
                                             norm=plt.Normalize(vmin=pressures.min(),
                                                                vmax=pressures.max()))
        for _, cell in frame.cells.items():
            color_to_fill = pressures_cb.to_rgba(float(cell.pressure))
            color_to_fill = mpl.colors.to_hex(color_to_fill)
            all_xs = [v.x for v in cell.get_cell_vertices()]
            all_ys = [v.y for v in cell.get_cell_vertices()]
            ax.fill(all_xs, all_ys, color_to_fill, alpha=0.4)

    plt.axis('off')
    if kwargs.get("mirror_y", False):
        plt.gca().invert_yaxis()
    if kwargs.get("mirror_x", False):
        plt.gca().invert_xaxis()

    if kwargs.get("colorbar", False):
        cb_stress = plt.colorbar(stress_cb, ax=ax)
        cb_stress.set_label("Stress [a.u.]")
        cb_stress.set_ticks([])
        if pressure:
            cb_pressure = plt.colorbar(pressures_cb, ax=ax)
            cb_pressure.set_label("Pressure [a.u.]")
            cb_pressure.set_ticks([])

    if kwargs.get("aspect_ratio", None):
        plt.gca().set_aspect(kwargs.get("aspect_ratio"))
    plt.tight_layout()
    return _, ax


def plot_difference(frame, folder: str, step: str, **kwargs) -> None:
    """Plot the difference between ground truth and inferred values in a heatmap.
    Both the inferred value and the ground truth are required

    :param frame: Frame to use
    :type frame: fframes.Frame
    :param folder: Folder to save plot
    :type folder: str
    :param step: File name of the plot
    :type step: str
    """
    plt.close()
    jet = plt.get_cmap('jet')
    _, ax = plt.subplots(1, 1)

    if not os.path.exists(folder):
        os.makedirs(folder)

    all_myosin = [edge.gt for edge in frame.edges.values()]
    all_myosin_max = max(all_myosin)

    for _, edge in frame.edges.items():
        difference_to_plot = abs(edge.tension - edge.gt) / all_myosin_max
        if difference_to_plot == 0:
            plt.plot(   (edge.v1.x, edge.v2.x),
                        (edge.v1.y, edge.v2.y),
                        color="black", linewidth=0.5, alpha=0.6)      

        else:
            color = jet(difference_to_plot)
            plt.plot(   (edge.v1.x, edge.v2.x),
                        (edge.v1.y, edge.v2.y),
                        color=color, linewidth=3)
    plt.axis('off')
    if kwargs.get("mirror_y", False):
        plt.gca().invert_yaxis()

    if kwargs.get("colorbar", False):
        sm = plt.cm.ScalarMappable(cmap="jet", norm=plt.Normalize(vmin=0, vmax=1))
        plt.colorbar(sm)

    name = os.path.join(folder, str(step) + ".png")
    plt.tight_layout()
    plt.savefig(name, dpi=500)
    plt.close()

def plot_force(freq: list, folder: str ='') -> None:
    """Plot histogram of the stresses.

    :param freq: List of stress values
    :type freq: list
    :param folder: Folder to save the log, defaults to ''
    :type folder: str, optional
    """
    plt.hist(freq, bins=25, density=True)
    plt.xlabel("Forces [ a. u ]")
    plt.ylabel("Frequency")
    plt.savefig(str(folder)+"log/forcesHist.png", dpi=500)
    plt.close()

def plot_mesh(frame: fframes.Frame,
              ax: mpl.axes.Axes = None,
              **kwargs) -> Tuple:
            #   xlim: list=[], ylim: list=[], 
            #   mirror_y: bool = False,
            #   mirror_x: bool = False) -> None:
    """Generate a plot of the mesh. Useful for visualizing IDs of cells, edges and vertices in the system.

    :param frame: Frame to plot
    :type frame: fframes.Frame
    :param xlim: Array of [x_min, x_max] to "zoom" in, defaults to []
    :type xlim: list, optional
    :param ylim: Array of [y_min, y_max] to "zoom" in, defaults to []
    :type ylim: list, optional
    :param mirror_y: If True the tissue is plotted as a mirror image in the Y axis, defaults to False
    :type mirror_y: bool, optional
    """
    if not ax:
        _, ax = plt.subplots(1, 1)
    if kwargs.get("plot_vertices", False) and not kwargs.get("plot_tjs", False):
        for v in frame.vertices.values():
            plt.scatter(v.x, v.y, s=2, color="black")
            if kwargs.get("plot_vertices_id", False):
                plt.annotate(str(v.id), [v.x, v.y], fontsize=2)

    if kwargs.get("plot_tjs", False):
        for v in frame.vertices.values():
            if len(v.ownEdges) >= 3:
                plt.scatter(v.x, v.y, s=2, color="black")
                if kwargs.get("plot_vertices_id", False):
                    plt.annotate(str(v.id), [v.x, v.y], fontsize=2)
    
    if kwargs.get("plot_versors", False):
        for v in frame.vertices.values():
            if len(v.ownEdges) >= 3:
                vertex_big_edges = [frame.big_edges[beid] for beid in v.own_big_edges]
                vertex_big_edges_versors = [big_edge.get_vector_from_vertex(v.id,
                                                                            fit_method="dlite") for big_edge in vertex_big_edges]
                # for versor in ve.get_versors(frame.vertices, frame.edges, v.id):
                for vector in vertex_big_edges_versors:
                    size = np.linalg.norm(vector)
                    versor = vector / size
                    plt.arrow(v.x,
                              v.y,
                              versor[0] * 10,
                              versor[1] * 10,
                              color="green",
                              width=4,
                              alpha=0.7,
                              zorder=1000)

    for e in frame.edges.values():
        if kwargs.get("plot_edges", False):
            plt.plot([e.v1.x, e.v2.x],
                     [e.v1.y, e.v2.y],
                     color="black",
                     linewidth=0.5,
                     alpha=0.6)

            if kwargs.get("plot_edges_id", False):
                plt.annotate(str(e.id), [(e.v1.x +  e.v2.x)/2 , (e.v1.y + e.v2.y)/2], fontweight="bold", fontsize=1)

    for c in frame.cells.values():
        cm = c.get_cm()
        cxs = [v.x for v in c.vertices]
        cys = [v.y for v in c.vertices]

        color_palette = kwargs.get("color_palette", "gray")
        if type(color_palette) is str:
            cell_colors = {k_id: color_palette for k_id in frame.cells.keys()}
        elif type(color_palette) is dict:
            cell_colors = color_palette
        else:
            raise ValueError("Color palette must be a string or a dictionary")
        # TODO add that if the id is not in the dictionary, it is colored in gray
        try:
            plt.fill(cxs, cys, alpha=0.25, color=cell_colors[c.id])
        except KeyError:
            plt.fill(cxs, cys, alpha=0.25, color="gray")
        if kwargs.get("plot_cells_id", False):
            plt.annotate(str(c.id), [cm[0], cm[1]], fontsize=5)

    xlim = kwargs.get("xlim", [])
    ylim = kwargs.get("ylim", [])
    if len(xlim) > 0:
        plt.xlim(xlim[0], xlim[1])
    if len(ylim) > 0:
        plt.ylim(ylim[0], ylim[1])
    if kwargs.get("mirror_y", False):
        plt.gca().invert_yaxis()
    if kwargs.get("mirror_x", False):
        plt.gca().invert_xaxis()

    plt.axis("off")
    return ax


def plot_equilibrium(mesh, step, folder, what="acceleration", normalized=False, cutoff=None):
    # According to how much acceleration is present, paint the tissue
    jet = plt.get_cmap('jet')

    # xcoords = [v.x for v in vertices.values()]
    # ycoords = [v.y for v in vertices.values()]

    # for v in vertices.values():
    #     plt.scatter(v.x, v.y, s=5,color="black")
        # plt.annotate(str(v.id), [v.x, v.y], fontsize=3)

    # for ev in mesh.mesh[step]['earr']:
    #     # calculate acceleration in both vertices of the edge
    #     acc0 = mesh.calculate_acceleration(ev[0], step)
    #     acc1 = mesh.calculate_acceleration(ev[-1], step)
    #     accelerations[mesh.mesh[step]['earr'].index(ev)] = np.linalg.norm(acc0) + np.linalg.norm(acc1)
    accelerations = mesh.whole_tissue_acceleration(step)
    maxAccel = np.nanmax(list(accelerations.values()))
    for ev in mesh.time_series[step].earr:
        for j in range(0, len(ev)):
            eBelong = mesh.time_series[step].vertices[ev[j]].ownEdges
            for eid in eBelong:
                edge = mesh.time_series[step].edges[eid]
                try:
                    if len(set(edge.get_vertices_id()) & set([ev[j], ev[j+1]])) == 2:
                        acceleration = accelerations[mesh.time_series[step].earr.index(ev)]
                        if normalized == "max":
                            color = jet(acceleration/maxAccel)
                        elif normalized == "cutoff":
                            color = "blue" if acceleration < cutoff else "red"
                        else:
                            color = jet(accelerations[mesh.time_series[step].earr.index(ev)])
                        
                        if np.isnan(acceleration):
                             plt.plot([edge.v1.x, edge.v2.x], [edge.v1.y, edge.v2.y], 
                                    color="gray", alpha=0.5, 
                                    linewidth=3)
                        else:
                            plt.plot([edge.v1.x, edge.v2.x], [edge.v1.y, edge.v2.y], 
                                    color=color, 
                                    linewidth=3)
                except IndexError:
                    pass
    
    if normalized == "max":
        vmin = 0
        vmax = maxAccel
    else:
        vmin = 0
        vmax = 1
    
    if normalized != "cutoff":
        sm = plt.cm.ScalarMappable(cmap="jet", norm=plt.Normalize(vmin=vmin, vmax=vmax))
        plt.colorbar(sm)
    plt.tight_layout()

    # plt.xlim()

    plt.savefig(os.path.join(folder, str(step) + ".png"), dpi=500)
    plt.clf()


def plot_time_connections(mesh: ftimes.TimeSeries, initial_time: int, final_time: int, folder: str='') -> None:
    """Plot time connections among vertices in all the frames between initial and final time

    :param mesh: TimeSeries object with mapping information
    :type mesh: ftimes.TimeSeries
    :param initial_time: Initial time
    :type initial_time: int
    :param final_time: Final time
    :type final_time: int
    :param folder: Path to save the plot, defaults to ''
    :type folder: str, optional
    """
    timerange = np.arange(initial_time, final_time-1, 1)
    for t in timerange:
        if mesh.mapping[t] != None:
            real_vertices_ids1 = np.array([[x[0]]+[x[-1]] for x in mesh.time_series[t].big_edges_list])
            real_vertices_ids2 = np.array([[x[0]]+[x[-1]] for x in mesh.time_series[t+1].big_edges_list])
            for v in mesh.time_series[t].vertices.values():
                if v.id in real_vertices_ids1 or v.id in mesh.time_series[t].border_vertices:
                    plt.scatter(v.x, v.y, s=5,color="black")
                    plt.annotate(str(v.id), [v.x, v.y], fontsize=4, color="black")
                    acceleration = mesh.calculate_displacement(v.id, t)

                    plt.arrow(v.x,
                              v.y, 
                              acceleration[0] * 1,
                              acceleration[1] * 1, color="red")
            

            for v in mesh.time_series[t+1].vertices.values():
                if v.id in real_vertices_ids2 or v.id in mesh.time_series[t+1].border_vertices:
                    plt.scatter(v.x, v.y, s=5, color="green")
                    plt.annotate(str(v.id), [v.x, v.y], fontsize=4, color="green")

            for e in mesh.time_series[t+1].edges.values():
                plt.plot([e.v1.x, e.v2.x], [e.v1.y, e.v2.y], color="green", alpha=0.2)
            for e in mesh.time_series[t].edges.values():
                plt.plot([e.v1.x, e.v2.x], [e.v1.y, e.v2.y], color="black", alpha=0.2)
                
            plt.savefig(os.path.join(folder, str(t)+".png"), dpi=500)
            plt.clf()

def plot_time_connections_two_times(mesh: ftimes.TimeSeries, t0: int, tf: int, 
                                    fname: str, folder: str="") -> None:
    """Create a connection plot among vertices, between only two timepoints

    :param mesh: TimeSeries object with mapping information
    :type mesh: ftimes.TimeSeries
    :param t0: initial time
    :type t0: int
    :param tf: Final time
    :type tf: int
    :param fname: File name to use
    :type fname: str
    :param folder: Folder to save the plot to, defaults to ""
    :type folder: str, optional
    """
    for v0 in mesh.time_series[t0].vertices.values():
        if True:
            v1_id = mesh.get_point_id_by_map(v0.id, t0, tf)
            v1 = mesh.time_series[tf].vertices[v1_id]

            plt.scatter(v0.x, v0.y, s=5, color="black")
            plt.scatter(v1.x, v1.y, s=5, color="green")

            deltax = v1.x - v0.x
            deltay = v1.y - v0.y

            plt.arrow(v0.x, v0.y, deltax, deltay, color="red")

    plt.axis("off")
    plt.tight_layout()
    plt.savefig(os.path.join(folder, str(fname)+".pdf"), dpi=500)
    plt.clf()



def plot_acceleration_heatmap(mesh: ftimes.TimeSeries, initial_time: int,
                              final_time: int, folder: str='', name: str='heatmap') -> None:
    """Generate an acceleration heatmap for each edge.

    :param mesh: TimeSeries object with mapping information
    :type mesh: ftimes.TimeSeries
    :param initial_time: Initial time
    :type initial_time: int
    :param final_time: Final time
    :type final_time: int
    :param folder: Folder to save plot to, defaults to ''
    :type folder: str, optional
    :param name: Filename for the plot, defaults to 'heatmap'
    :type name: str, optional
    """
    t0 = mesh.time_series[initial_time]
    heatmap = []
    for ii in range(0, len(t0.earr)):
        row = mesh.acceleration_per_edge(ii, initial_time, final_time)
        if not np.all(np.isnan(row)):
            heatmap.append(row) 
    
    save_heatmap(heatmap, folder, initial_time, final_time, name=name)

def get_velocity_heatmap(mesh: ftimes.TimeSeries, initial_time: int, final_time: int) -> list:
    """Generate an array with the velocity of each edge tracked through time.

    :param mesh: TimeSeries object with mapping information
    :type mesh: ftimes.TimeSeries
    :param initial_time: Initial time
    :type initial_time: int
    :param final_time: Final time
    :type final_time: int
    :return: Array where each row is the velocity of an edge and each column a different time
    :rtype: list
    """
    t0 = mesh.time_series[initial_time]
    heatmap = []
    for ii in range(0, len(t0.earr)):
        row = mesh.velocity_per_edge(ii, initial_time, final_time)
        if not np.all(np.isnan(row)):
            heatmap.append(row)
    return heatmap

def save_heatmap(heatmap: list, folder: str, initial_time: int, final_time:int, name: str='heatmap') -> None:
    """Save a given heatmap

    :param heatmap: Heatmap array
    :type heatmap: list
    :param folder: Folder to save the plot to
    :type folder: str
    :param initial_time: Initial time
    :type initial_time: int
    :param final_time: Final time
    :type final_time: int
    :param name: Name of the plot, defaults to 'heatmap'
    :type name: str, optional
    """
    plt.imshow(heatmap)
    plt.colorbar()
    plt.xticks(np.arange(0, final_time - initial_time, 5), np.arange(initial_time, final_time, 5))
    plt.tight_layout()
    plt.savefig(os.path.join(folder, str(name)+"_"+str(initial_time)+".png"), dpi=500)
    plt.clf()

def plot_ablated_edge(frames: dict, step: int, bedge_index: int, folder: str) -> None:
    """Plot the mesh of the system with the ablated edges in the tissue in a different color.

    :param frames: Dictionary with all the frames
    :type frames: dict
    :param step: Step to plot. Must correspond with a time in the frames dictionary
    :type step: int
    :param bedge_index: Index of the ablated big edge
    :type bedge_index: int
    :param folder: Folder to save the mesh
    :type folder: str
    """
    # TODO: integrate with  plot_mesh() as an option
    for v in frames[step].vertices.values():
        plt.scatter(v.x, v.y, s=2, color="black")
        plt.annotate(str(v.id), [v.x, v.y], fontsize=2)

    for e in frames[step].edges.values():
        plt.plot([e.v1.x, e.v2.x], [e.v1.y, e.v2.y], color="orange", linewidth=0.5)

    for v in frames[step].earr[bedge_index]:
        for eid in frames[step].vertices[v].ownEdges:
            e = frames[step].edges[eid]
            plt.plot([e.v1.x, e.v2.x], [e.v1.y, e.v2.y], color="green", linewidth=1.5)

    for c in frames[step].cells.values():
        cm = c.get_cm()
        cxs = [v.x for v in c.vertices]
        cys = [v.y for v in c.vertices]

        plt.fill(cxs, cys, alpha=0.3)
        plt.annotate(str(c.id), [cm[0], cm[1]], fontsize=5)

    plt.savefig(folder, dpi=600)
    plt.close()


def plot_residues(frame: fframes.Frame, folder: str, normalized: bool=False, maxForce:float=None, 
                    minForce: float=None, mirror_y: bool=False) -> None:
    """Plot differences between inferred values and ground truths as residue
    vectors from each pivot vertex

    :param frame: Frame with all infromation
    :type frame: fframes.Frame
    :param folder: Folder to save the plot
    :type folder: str
    :param normalized: "max" normalizes by the maximum force; "normal" normalizes to a \
        normal distribution centered in zero, defaults to False
    :type normalized: bool, optional
    :param maxForce:  Maximum force to use in the normalization , defaults to None
    :type maxForce: float, optional
    :param minForce:  Minimum force to use in the normalization, defaults to None
    :type minForce: float, optional
    :param mirror_y: Plot mirroring the y component, defaults to False
    :type mirror_y: bool, optional
    """
    plt.close()
    jet = plt.get_cmap('jet')
    _, ax = plt.subplots(1,1)

    if not os.path.exists(folder):
        os.makedirs(folder)
    
    all_forces = [edge.tension for edge in frame.edges.values() if edge.tension != 0]

    if maxForce == None:
        maxForce = max(all_forces)
    if minForce == None:
        minForce = min(all_forces)

    allValues = all_forces
    aveForce = np.mean(allValues)
    stdForce = np.std(allValues)

    for _, edge in frame.edges.items():
        if edge.tension == 0:
            plt.plot(   (edge.v1.x, edge.v2.x),
                        (edge.v1.y, edge.v2.y),
                        color="black", linewidth=0.5, alpha=0.6)
        else:
            if normalized == "normal":
                forceValToPlot = (edge.tension - aveForce) / stdForce
            elif normalized == "max":
                forceValToPlot = edge.tension / maxForce
            else:
                forceValToPlot = edge.tension/maxForce
            
            color = jet(forceValToPlot)
            plt.plot(   (edge.v1.x, edge.v2.x),
                        (edge.v1.y, edge.v2.y),
                        color=color, linewidth=3)
    
    tj_vertices = list(set([big_edge[0] for big_edge in frame.big_edges_list]))

    for vid in tj_vertices:
        force_difference = np.array([0, 0]).astype(np.float64)
        for versor in ve.get_versors(frame.vertices, frame.edges, vid):
            try:
                calculated_force_vector = [frame.edges[versor[4]].tension * versor[0], 
                                            frame.edges[versor[4]].tension * versor[1]]
                original_force_vector = [frame.edges[versor[4]].gt * versor[0],
                                        frame.edges[versor[4]].gt * versor[1]]
            except KeyError:
                raise("Missing tension or ground truth")
            
            force_difference += np.array(calculated_force_vector) - \
                                np.array(original_force_vector)

        plt.arrow(frame.vertices[vid].x, frame.vertices[vid].y, force_difference[0],
                    force_difference[0], color="black", zorder=100)

    plt.axis('off')
    if mirror_y:
        plt.gca().invert_yaxis()
    
    name = os.path.join(folder, str(frame.time) + ".png")
    plt.tight_layout()
    plt.savefig(name, dpi=500)
    plt.close()
    
def plot_stress_tensor(frame: fframes.Frame, folder: str, fname: str, 
                       grid: int=5, radius: float=1, **kwargs) -> None:
    """Plot the principal components of the stress tensor on top of a 
    layout of the mesh

    :param frame: Frame object
    :type frame: fframes.Frame
    :param folder: Folder to save the output plot
    :type folder: str
    :param fname: Name for the plot
    :type fname: str
    :param grid: Number of grid points in each axis, defaults to 5
    :type grid: int, optional
    :param radius: Number of average cell radii to use in the integration, defaults to 1
    :type radius: float, optional
    """
    frame.calculate_stress_tensor(grid, radius)
    plt.close()
    _, ax = plt.subplots(1,1)

    if not os.path.exists(folder):
        os.makedirs(folder)
    
    for _, edge in frame.edges.items():
        plt.plot(   (edge.v1.x, edge.v2.x),
                    (edge.v1.y, edge.v2.y),
                    color="black", linewidth=0.5, alpha=0.6)

    external_scale = kwargs.get("tensor_scale", 1)
    try:
        scale_factor = (external_scale/10) * min(frame.stress_tensor[1][0][0] - frame.stress_tensor[1][0][1],
                                                frame.stress_tensor[1][1][0] - frame.stress_tensor[1][1][1])
    except IndexError:
        scale_factor = (external_scale/10) * 0.1 * min(frame.stress_tensor[1][0][0],
                                                       frame.stress_tensor[1][1][0] )

    for positions, eigensystem in frame.principal_stress.items():
        eigen_norm = np.linalg.norm(eigensystem[0])
        if eigen_norm == 0:
            rescaled_eigenvalues = [0, 0]
        else:
            rescaled_eigenvalues = eigensystem[0] / eigen_norm
        component1 = scale_factor * rescaled_eigenvalues[0] * eigensystem[1][:, 0]
        component2 = scale_factor * rescaled_eigenvalues[1] * eigensystem[1][:, 1]

        axis_1_x = [positions[0] - component1[0], positions[0], positions[0] + component1[0]]
        axis_1_y = [positions[1] - component1[1], positions[1], positions[1] + component1[1]]
        axis_2_x = [positions[0] - component2[0], positions[0], positions[0] + component2[0]]
        axis_2_y = [positions[1] - component2[1], positions[1], positions[1] + component2[1]]

        plt.plot(axis_1_x, axis_1_y, color="green", linewidth=2)
        plt.plot(axis_2_x, axis_2_y, color="green", linewidth=2)

    plt.axis('off')

    if kwargs.get("mirror_y", None):
        plt.gca().invert_yaxis()


    name = os.path.join(folder, str(fname))
    plt.tight_layout()
    plt.savefig(name, dpi=500)
    plt.close()


def plot_skeleton(frame: fframes.Frame, save_folder: str, **kwargs) -> None:

    maximize = kwargs.get("maximize", False)
    if maximize:
        factor_x, factor_y = maximize
    else:
        factor_x, factor_y = 1, 1

    new_frame = deepcopy(frame)
    for _, vertex in new_frame.vertices.items():
        vertex.x = vertex.x * factor_x
        vertex.y = vertex.y * factor_y

    for beid, big_edge in new_frame.big_edges.items():
        new_frame.big_edges[beid] = fedges.BigEdge(beid,
                                                   big_edge.vertices)
    frame = new_frame

    if kwargs.get("use_all", False):
        edges_to_use = list(frame.big_edges.values())
    else:
        edges_to_use = frame.internal_big_edges

    all_vertices = set()
    for _, big_edge in enumerate(edges_to_use):
        vertices, _ = fmyo.get_interpolation(big_edge, layers=0)
        all_vertices.update(vertices)

    all_vertices_tuple = list(zip(*all_vertices))

    all_xs = all_vertices_tuple[0]
    all_ys = all_vertices_tuple[1]
    max_x, min_x = np.max(all_xs), np.min(all_xs)
    max_y, min_y = np.max(all_ys), np.min(all_ys)

    size_x = int(max_x - min_x)
    size_y = int(max_y - min_y)

    image_array = np.zeros((size_y + 10, size_x + 10))

    for vx, vy in list(all_vertices):
        vx = int((vx - min_x)) + 5
        vy = int((vy - min_y)) + 5

        image_array[int(vy), int(vx)] = 255
    im = PIL.Image.fromarray(image_array)
    im.save(save_folder)


def plot_inference_as_tiff(forsys: fs.ForSys,
                           image_sizes: list,
                           save_folder: str,
                           **kwargs) -> None:
    """Generate a plot of a tissue using the inferred tensions and pressures
        and save it as a tiff. If multiple more than one frame exists, the
        image is saved as a stack

    :param forsys: ForSys object with all the frames already solved
    :type frame: fs.ForSys
    :param pressure: If True pressures are plotted in the inner space of the cells.
    :type pressure: bool
    """
    stack = []
    original_images = kwargs.get("original_images", [])
    for time, frame in forsys.frames.items():
        original_size = image_sizes[time]
        image_array = np.zeros((original_size[1], original_size[0]))

        if kwargs.get("use_all", True):
            edges_to_use = list(frame.big_edges.values())
        else:
            edges_to_use = frame.internal_big_edges

        # from 0 to 255
        max_tension = np.max([big_edge.tension for big_edge in edges_to_use])

        for _, big_edge in enumerate(edges_to_use):
            vertices, _ = fmyo.get_interpolation(big_edge,
                                                 layers=kwargs.get("layers",
                                                                   0))
            for vx, vy in vertices:
                if big_edge.external:
                    image_array[int(vy), int(vx)] = kwargs.get("external_color", 255)
                else:
                    image_array[int(vy), int(vx)] = (big_edge.tension / max_tension) * 255
        if original_images:
            original_image = original_images[time]
            image_array_original = np.array(original_image)
            complete_frame = np.array([image_array_original, image_array])
        else:
            complete_frame = np.array([image_array])
        stack.append(complete_frame)

    stack = np.array(stack).astype(np.uint8)
    print(f"Complete size: {stack.shape}")
    tif.imwrite(save_folder,
                stack,
                imagej=True,
                # TODO: add resolution from original image
                # resolution=(1, 1),
                metadata={
                        "mode": "composite",
                        "axes": "TCYX",
                        "Channel": {"Name": ["Microscopy", "Stress"]}
                },
                )


def plot_big_edges(frame, **kwargs) -> Tuple:
    fig, ax = plt.subplots(1, 1)

    for v in frame.vertices.values():
        if len(v.ownEdges) > 2:
            plt.scatter(v.x, v.y, s=2, color="black")
            plt.annotate(str(v.id), [v.x, v.y], fontsize=2)

    for e in frame.edges.values():
        plt.plot([e.v1.x, e.v2.x],
                 [e.v1.y, e.v2.y], color="orange", linewidth=0.5)

    for be in frame.big_edges.values():
        x_cm = np.mean(be.xs)
        y_cm = np.mean(be.ys)
        plt.annotate(str(be.big_edge_id), [x_cm , y_cm], fontweight="bold", fontsize=5, color="blue", alpha=0.4)

    for c in frame.cells.values():
        cm = c.get_cm()
        cxs = [v.x for v in c.vertices]
        cys = [v.y for v in c.vertices]

        plt.fill(cxs, cys, alpha=0.3)
        plt.annotate(str(c.id), [cm[0], cm[1]])

    kwarg_xlim = kwargs.get("xlim", [])
    kwarg_ylim = kwargs.get("ylim", [])

    if len(kwarg_xlim) > 0:
        plt.xlim(kwarg_xlim[0], kwarg_xlim[1])
    if len(kwarg_ylim) > 0:
        plt.ylim(kwarg_ylim[0], kwarg_ylim[1])
    if kwargs.get("mirror_y", False):
        plt.gca().invert_yaxis()
    if kwargs.get("mirror_x", False):
        plt.gca().invert_xaxis()

    return fig, ax



from dataclasses import dataclass, field
import pandas as pd
import numpy as np
import forsys.time_series as ts
import forsys.fmatrix as fmatrix
import forsys.pmatrix as pmatrix
import forsys.frames as fsframes

@dataclass
class ForSys():
    """
    Class wrapper containing the interface with the ForSys solver

    :param frames: Unique identifier of the cell
    :type id: dict
    :param cm: True if the each frame should be moved the the center of mass \
        system, defaults to False
    :type cm: bool
    :param initial_guess: Dictionary pairing vertex IDs in different timepoints \
        to force ForSys to connect them through time
    :type initial_guess: list
    """
    frames: dict

    cm: bool = False
    initial_guess: list = field(default_factory=list)


    def __post_init__(self):
        """Constructor method        
        """
        # if there is more than 1 time, generate time series
        if len(self.frames) > 1:
            self.mesh = ts.TimeSeries(self.frames, cm=self.cm, 
                                        initial_guess=self.initial_guess)
            
            self.times_to_use = self.mesh.times_to_use()
        else:
            self.mesh = {}
        
        self.force_matrices = {}
        self.pressure_matrices = {}
        self.forces = {k: None for k in range(len(self.frames))}
        self.pressures = {k: None for k in range(len(self.frames))}

    
    def build_force_matrix(self, when: int = 0, term: str = "none", metadata:dict = {}, **kwargs) -> None:
        """
        Interface to create the matrix system to solve for the stresses

        :param when: Time at which the matrix should be constructed, defaults to 0
        :type when: int, optional
        :param term: Define behaviour for external vertices. By default they are not \
            considered, defaults to "none"
        :type term: str, optional
        :param metadata: Extra arguments to supply to the matrix builder, defaults to {}
        :type metadata: dict, optional
        """
        # TODO: add matrix caching
        self.force_matrices[when] = fmatrix.ForceMatrix(self.frames[when],
                                # externals_to_use = self.frames[when].border_vertices,
                                externals_to_use = 'none',
                                term=term,
                                metadata=metadata,
                                timeseries=self.mesh,
                                angle_limit=kwargs.get("angle_limit", np.pi),
                                circle_fit_method=kwargs.get("circle_fit_method", "dlite"),)


    def build_pressure_matrix(self, when: int = 0):
        """
        Interface to create the matrix system to solve for the pressures. 
        Must be called after stresses were already found.

        :param when: Time at which the matrix should be constructed, defaults to 0
        :type when: int, optional
        """
        # TODO: add matrix caching
        self.pressure_matrices[when] = pmatrix.PressureMatrix(self.frames[when],
                                                              timeseries=self.mesh)
        

    def solve_stress(self, when: int = 0, **kwargs) -> None:
        """
        Interface to call the solver method for the stresses.
        The matrix must already exist.

        :param when: Time at which to calculate the solution, defaults to 0
        :type when: int, optional
        """
        self.forces[when] = self.force_matrices[when].solve(self.mesh, **kwargs)
        self.frames[when].forces = self.forces[when]
        self.frames[when].assign_tensions_to_big_edges()


    def solve_pressure(self, when: int=0, **kwargs):
        """Interface to call the solver method for the pressures.
        The matrix must already exist.

        :param when: Time at which to calculate the solution, defaults to 0
        :type when: int, optional
        """
        assert type(self.forces[when]) is not None, "Forces must be calculated first"
        self.pressures = self.pressure_matrices[when].solve_system(**kwargs)
        self.frames[when].assign_pressures(self.pressures, self.pressure_matrices[when].mapping_order)


    def log_force(self, when: int) -> pd.DataFrame:
        """Create dataframe with the solution for the stresses at a 
        given time.

        :param when: Time at which to create the dataframe
        :type when: int
        :return: Dataframe with the stresses per edge
        :rtype: pd.DataFrame
        """
        df = pd.DataFrame.from_dict(self.frames[when].forces, orient='index')
        df['is_border'] = [False if isinstance(x, int) else True for x in df.index]
        return df


    def get_edge_force(self, v0: int, v1: int, t0: int = -1, tmax: int = -1) -> list:
        """
        Return the edge tension between two timepoints for an edge, given two vertices.
        If no time is provided, the whole evolution is assumed.

        :param v0: First vertex of the edge
        :type v0: int
        :param v1: Second vertex of the edge
        :type v1: int
        :param t0: Initial time at which the edge's stress is required, defaults to -1. \
            Default returns the value for each frame.
        :type t0: int, optional
        :param tmax: Final time for the edge's stress, defaults to -1.
        :type tmax: int, optional
        :return: Array with the values of the edge's stress in time
        :rtype: list
        """
        edge_force = []
        if tmax == -1 or t0 == -1:
            range_values = self.mesh.mapping.keys()
        else:
            range_values = range(t0, tmax)

        for t in range_values:
            current_v0 = self.mesh.get_point_id_by_map(v0, t0, t)
            current_v1 = self.mesh.get_point_id_by_map(v1, t0, t)
            for e in self.frames[t].earr:
                if current_v0 in e and current_v1 in e:
                    edge_id = self.frames[t].earr.index(e)
            edge_force.append(self.frames[t].forces[edge_id])

        return edge_force          
            
            
    def remove_outermost_edges(self, frame_number: int, layers: int = 1) -> tuple:
        """Remove outermost layer of cells from a tissue. 
        WARNING: This function is experimental

        :param frame_number: Frame to perform the removal at
        :type frame_number: int
        :param layers: Number of layers of cells to remove, defaults to 1
        :type layers: int, optional
        :return: New vertices, edges and cells dictionaries after removal
        :rtype: tuple
        """
        if layers == 0:
            return True
        elif layers > 1:
            raise(NotImplementedError)
        else:
            # vertex_to_delete = []
            border_cells_ids = [cell.id for cell in self.frames[0].cells.values() if cell.is_border]
            for cell_id in border_cells_ids:
                self.remove_cell(frame_number, cell_id)            
            
            return self.frames[frame_number].vertices, self.frames[frame_number].edges, self.frames[frame_number].cells


    def remove_cell(self, frame_number: int, cell_id: int) -> fsframes.Frame:
        """Remove a cell from the tissue

        :param frame_number: Timepoint to perform removal
        :type frame_number: int
        :param cell_id: ID of the cell to be removed
        :type cell_id: int
        :return: New state of the frame
        :rtype: fsframes.Frame
        """
        vertex_to_delete = []
        for vertex in self.frames[frame_number].cells[cell_id].vertices:
            if len(vertex.ownCells) < 2:
                assert vertex.ownCells[0] == self.frames[frame_number].cells[cell_id].id, "Vertex incorrectly placed in cell"
                for ii in vertex.ownEdges.copy():
                    del self.frames[frame_number].edges[ii]

                vertex_to_delete.append(vertex.id)
        
        for vertex_id in list(set(vertex_to_delete)):
            del self.frames[frame_number].vertices[vertex_id]

        del self.frames[0].cells[cell_id]

        self.frames[frame_number] = fsframes.Frame(frame_number, 
                                                   self.frames[frame_number].vertices, 
                                                   self.frames[frame_number].edges, 
                                                   self.frames[frame_number].cells, 
                                                   time=self.frames[frame_number].time,
                                                   gt=self.frames[frame_number].gt)
        return self.frames[frame_number]


    def get_system_velocity_per_frame(self, time_interval: list=None, angle_limit=np.inf) -> list:
        """Calculate the average velocity of the frame. Useful for 
        normalization purposes. When this functions is called, the system's
        matrix is automatically generated.

        :param time_interval: Times to do the calculation. Default is every frame, defaults to None
        :type time_interval: list, optional
        :param angle_limit: Angle limit to ignore triple junctions in the inference, defaults to np.inf
        :type angle_limit: _type_, optional
        :return: List of velocity averages per frame
        :rtype: list
        """
        velocities_per_frame = []
        if type(time_interval) is not list:
            time_interval = range(len(self.frames))
        for time in time_interval:
            self.build_force_matrix(when=time,
                                    angle_limit=angle_limit)
            _, ave_velocity = self.force_matrices[time].set_velocity_matrix(self.mesh,
                                                                            b_matrix="velocity",
                                                                            adimensional_velocity=True)
            velocities_per_frame.append(ave_velocity)

        return velocities_per_frame


import pytest
import os

import forsys as fs


@pytest.fixture
def cells_3_orient1():
    filename = os.path.join("tests", "data", "skeleton", "3_no_cells.npy")
    skeleton = fs.skeleton.Skeleton(filename,
                                    minimum_distance=10)
    vertices, edges, cells = skeleton.create_lattice()
    frame = fs.frames.Frame(0,
                            vertices,
                            edges,
                            cells)
    yield frame


@pytest.fixture
def triangle_holes():
    filename = os.path.join("tests", "data", "skeleton", "triangle_holes.npy")
    skeleton = fs.skeleton.Skeleton(filename,
                                    minimum_distance=10)
    vertices, edges, cells = skeleton.create_lattice()
    frame = fs.frames.Frame(0,
                            vertices,
                            edges,
                            cells)
    yield frame


def test_3_orient1(cells_3_orient1):
    all_edges = {tuple(sorted(e.get_vertices_id())): e.id
                 for e in cells_3_orient1.edges.values()}
    assert len(cells_3_orient1.vertices[368].ownCells) == 2
    assert len(cells_3_orient1.vertices[368].ownEdges) == 3

    assert tuple(sorted([350, 1764])) not in all_edges
    assert tuple(sorted([1909, 2131])) not in all_edges
    assert tuple(sorted([2071, 1209])) not in all_edges


def test_triangle_holes(triangle_holes):
    assert len(triangle_holes.vertices[1541].ownEdges) == 2
    assert len(triangle_holes.vertices[766].ownEdges) == 3

    assert len(triangle_holes.vertices[2243].ownEdges) == 3
    assert len(triangle_holes.vertices[2243].ownCells) == 3


def test_wrong_format():
    filename = os.path.join("tests", "data", "skeleton", "wrong_format.format")
    with pytest.raises(ValueError):
        fs.skeleton.Skeleton(filename,
                             minimum_distance=0)



import pytest
import os 

import numpy as np
import forsys as fs
import warnings
from sklearn.metrics import mean_absolute_percentage_error, r2_score

@pytest.fixture
def furrow():
    frames = {}
    for ii in range(0, 8):
        surfaceEvolver = fs.surface_evolver.SurfaceEvolver(os.path.join(
                                                            "tests", 
                                                            "data", 
                                                            "furrow_gauss_velocity", 
                                                            f"stage{ii}.dmp"))
        frames[ii] = fs.frames.Frame(ii,
                                    surfaceEvolver.vertices,
                                    surfaceEvolver.edges, 
                                    surfaceEvolver.cells, 
                                    time=ii, 
                                    gt=True)
    forsys = fs.ForSys(frames, cm=False)
    yield forsys

def test_fit_initial_furrow(furrow):
    furrow.build_force_matrix(when=0)
    furrow.solve_stress(when=0)

    tensions_df = furrow.frames[0].get_tensions()

    r_value = r2_score(tensions_df['gt'].values / tensions_df['gt'].mean(), 
                                tensions_df['stress'].values)
    print("initial furrow", r_value)
    assert 1 > r_value > 0.94


def test_fit_last_furrow(furrow):
    last_frame = len(furrow.frames) - 1
    furrow.build_force_matrix(when=last_frame)
    furrow.solve_stress(when=last_frame)
    
    tensions_df = furrow.frames[last_frame].get_tensions()

    r_value = r2_score(tensions_df['gt'].values / tensions_df['gt'].mean(), 
                                tensions_df['stress'].values)
    print("Final furrow", r_value)
    assert 1 > r_value > 0.93


def test_fit_last_furrow_taubin(furrow):
    last_frame = len(furrow.frames) - 1
    furrow.build_force_matrix(when=last_frame, circle_fit_method="taubinSVD")
    furrow.solve_stress(when=last_frame)
    
    tensions_df = furrow.frames[last_frame].get_tensions()

    r_value = r2_score(tensions_df['gt'].values / tensions_df['gt'].mean(), 
                                tensions_df['stress'].values)
    print("Final furrow", r_value)
    assert 1 > r_value > 0.93


def test_fit_furrow_movement_velocity(furrow):
    all_r_values = []
    for ii in furrow.frames.keys():
        furrow.build_force_matrix(when=ii)
        furrow.solve_stress(when=ii, b_matrix="velocity")
        tensions_df = furrow.frames[ii].get_tensions()

        arguments = [tensions_df['gt'].values / tensions_df['gt'].mean(), 
                                    tensions_df['stress'].values / tensions_df['stress'].mean()]
        r_value = r2_score(*arguments)
        print(f"t ={ii} with MAPE {mean_absolute_percentage_error(*arguments)}")
        all_r_values.append(r_value)
    print("Dynamic furrow", all_r_values)
    assert np.all([1 > value > 0.94 for value in all_r_values])

def test_fit_furrow_movement_velocity_taubin(furrow):
    all_r_values = []
    for ii in furrow.frames.keys():
        furrow.build_force_matrix(when=ii, circle_fit_method="taubinSVD")
        furrow.solve_stress(when=ii, b_matrix="velocity")
        tensions_df = furrow.frames[ii].get_tensions()

        arguments = [tensions_df['gt'].values / tensions_df['gt'].mean(), 
                                    tensions_df['stress'].values / tensions_df['stress'].mean()]
        r_value = r2_score(*arguments)
        print(f"t ={ii} with MAPE {mean_absolute_percentage_error(*arguments)}")
        all_r_values.append(r_value)
    print("Dynamic furrow", all_r_values)
    assert np.all([1 > value > 0.94 for value in all_r_values])

def test_fit_furrow_velocity_lsq_initial_last_val(furrow, recwarn):
    all_r_values = []
    for ii in furrow.frames.keys():
        if ii == 0:
            initial_condition = np.ones(len(furrow.frames[0].internal_big_edges))
        else:
            initial_condition = furrow.frames[ii-1].get_tensions()["stress"]

        furrow.build_force_matrix(when=ii)
        furrow.solve_stress(when=ii,
                            b_matrix="velocity",
                            initial_condition=initial_condition,
                            method="lsq")            
            
        tensions_df = furrow.frames[ii].get_tensions()
        r_value = r2_score(tensions_df['gt'].values / tensions_df['gt'].mean(), 
                                    tensions_df['stress'].values / tensions_df['stress'].mean())
        all_r_values.append(r_value)
    print("Dynamic furrow using last", all_r_values)
    for r in recwarn:
        assert "Numerically" not in str(r.message), "Warnings raised while solving, probably LSQ went to numerical"
    assert np.all([1 > value > 0.92 for value in all_r_values])


def test_fit_furrow_velocity_lsq_initial_gt(furrow, recwarn):
    all_r_values = []
    for ii in furrow.frames.keys():
        initial_condition = list(furrow.frames[ii].get_gt_tensions()["gt"].values / furrow.frames[ii].get_gt_tensions()["gt"].mean())

        furrow.build_force_matrix(when=ii)
        furrow.solve_stress(when=ii,
                            b_matrix="velocity",
                            initial_condition=initial_condition,
                            method="lsq")
        tensions_df = furrow.frames[ii].get_tensions()
        r_value = r2_score(tensions_df['gt'].values / tensions_df['gt'].mean(), 
                                    tensions_df['stress'].values / tensions_df['stress'].mean())
        all_r_values.append(r_value)
    print("Dynamic furrow using GT", all_r_values)
    print(tensions_df)
    for r in recwarn:
        assert "Numerically" not in str(r.message), "Warnings raised while solving, probably LSQ went to numerical"
    assert np.all([1 > value > 0.94 for value in all_r_values])

def test_fit_furrow_velocity_lsq_initial_one(furrow, recwarn):
    all_r_values = []
    for ii in furrow.frames.keys():
        initial_condition = np.ones(len(furrow.frames[ii].get_gt_tensions()["gt"]))
        # initial_condition = np.abs(np.random.normal(1, 0.1, len(furrow.frames[ii].get_gt_tensions()["gt"])))
        

        furrow.build_force_matrix(when=ii)
        furrow.solve_stress(when=ii,
                            b_matrix="velocity",
                            initial_condition=initial_condition,
                            method="lsq")
        tensions_df = furrow.frames[ii].get_tensions()
        # r_value = distance_to_yeqx(tensions_df['gt'].values / tensions_df['gt'].mean(), 
        #                             tensions_df['stress'].values / tensions_df['stress'].mean())
        arguments = [tensions_df['gt'].values / tensions_df['gt'].mean(), 
                                    tensions_df['stress'].values / tensions_df['stress'].mean()]
        r_value = r2_score(*arguments)
        print(f"t ={ii} with MAPE {mean_absolute_percentage_error(*arguments)}")
        all_r_values.append(r_value)
    print("Dynamic furrow using Ones", all_r_values)
    print(tensions_df)
    for r in recwarn:
        assert "Numerically" not in str(r.message), "Warnings raised while solving, probably LSQ went to numerical"
    assert np.all([1 > value > 0.95 for value in all_r_values])

@pytest.mark.skip(reason="Eliminating edges doesn't work with this normalization ")
def test_fit_furrow_velocity_lsq_initial_gt_deleting10(furrow, recwarn):
    all_r_values = []
    angle_limit = 180 * (1 - 0.1) * np.pi / 180
    for ii in furrow.frames.keys():
        initial_condition = list(furrow.frames[ii].get_gt_tensions()["gt"].values / furrow.frames[ii].get_gt_tensions()["gt"].mean())

        furrow.build_force_matrix(when=ii,
                                  angle_limit=angle_limit)
        furrow.solve_stress(when=ii,
                            b_matrix="velocity",
                            initial_condition=initial_condition,
                            method="lsq")
        tensions_df = furrow.frames[ii].get_tensions()
        
        arguments = [tensions_df['gt'].values / tensions_df['gt'].mean(), 
                                    tensions_df['stress'].values / tensions_df['stress'].mean()]
        r_value = r2_score(*arguments)
        all_r_values.append(r_value)
        print(f"t ={ii} with MAPE {mean_absolute_percentage_error(*arguments)}")
    print("Dynamic furrow using GT with 10% discard", all_r_values)
    for r in recwarn:
        assert "Numerically" not in str(r.message), "Warnings raised while solving, probably LSQ went to numerical"
    assert np.all([1 > value > 0.94 for value in all_r_values])


def test_fit_furrow_velocity_lsq_initial_one_std(furrow, recwarn):
    all_r_values = []
    for ii in furrow.frames.keys():
        initial_condition = np.ones(len(furrow.frames[ii].get_gt_tensions()["gt"]))
        # initial_condition = np.abs(np.random.normal(1, 0.1, len(furrow.frames[ii].get_gt_tensions()["gt"])))
        

        furrow.build_force_matrix(when=ii)
        furrow.solve_stress(when=ii,
                            b_matrix="velocity",
                            initial_condition=initial_condition,
                            method="lsq",
                            use_std=True)
        tensions_df = furrow.frames[ii].get_tensions()
        # r_value = distance_to_yeqx(tensions_df['gt'].values / tensions_df['gt'].mean(), 
        #                             tensions_df['stress'].values / tensions_df['stress'].mean())
        arguments = [tensions_df['gt'].values / tensions_df['gt'].mean(), 
                                    tensions_df['stress'].values / tensions_df['stress'].mean()]
        r_value = r2_score(*arguments)
        print(f"t ={ii} with MAPE {mean_absolute_percentage_error(*arguments)}")
        all_r_values.append(r_value)
    print("Dynamic furrow using Ones", all_r_values)
    print(tensions_df)
    for r in recwarn:
        assert "Numerically" not in str(r.message), "Warnings raised while solving, probably LSQ went to numerical"
    assert np.all([1 > value > 0.95 for value in all_r_values])



import pytest

import numpy as np
import forsys as fs


@pytest.fixture
def set_up_mesh():
    frames = {}
    vertices = {}
    edges = {}
    cells = {}

    vertices[0] = fs.vertex.Vertex(0, 0, 1)
    vertices[1] = fs.vertex.Vertex(1, 0, 1)
    edges[0] = fs.edge.SmallEdge(0, vertices[0], vertices[1])
    cells[0] = fs.cell.Cell(0, vertices.values())
    frames[0] = fs.frames.Frame(0, vertices, edges, cells, time=0)

    vertices[0] = fs.vertex.Vertex(0, 4, 4)
    vertices[1] = fs.vertex.Vertex(1, 0, 1)
    edges[0] = fs.edge.SmallEdge(0, vertices[0], vertices[1])
    cells[0] = fs.cell.Cell(0, vertices.values())
    frames[1] = fs.frames.Frame(1, vertices, edges, cells, time=1)

    vertices[0] = fs.vertex.Vertex(0, 12, 11)
    vertices[1] = fs.vertex.Vertex(1, 0, 1)
    edges[0] = fs.edge.SmallEdge(0, vertices[0], vertices[1])
    cells[0] =  fs.cell.Cell(0, vertices.values())
    frames[2] = fs.frames.Frame(2, vertices, edges, cells, time=2)
    
    vertices[0] = fs.vertex.Vertex(0, 18, 22)
    vertices[1] = fs.vertex.Vertex(1, 0, 1)
    edges[0] = fs.edge.SmallEdge(0, vertices[0], vertices[1])
    cells[0] = fs.cell.Cell(0, vertices.values())
    frames[3] = fs.frames.Frame(3, vertices, edges, cells, time=3)

    forsys = fs.ForSys(frames, cm=False)
    forsys.mesh.mapping = {0: {0: 0, 1:1}, 2: {0: 0, 1:1}, 1: {0: 0, 1:1}}

    yield forsys.mesh
    
    del forsys
    del vertices
    del edges
    del cells

@pytest.mark.skip(reason="Cell neighbor calculation doesn't work with this test")
def test_acceleration(set_up_mesh):
    assert np.all(set_up_mesh.calculate_acceleration(0, 0) == [4., 4.])
    assert np.all(set_up_mesh.calculate_acceleration(0, 1) == [4., 4.])
    assert np.all(set_up_mesh.calculate_acceleration(0, 2) == [-2., 4.])
    assert np.all(set_up_mesh.calculate_acceleration(0, 3) == [-2., 4.])

@pytest.mark.skip(reason="Cell neighbor calculation doesn't work with this test")
def test_velocity(set_up_mesh):
    assert np.all(set_up_mesh.calculate_velocity(0, 0) == [4., 3.])
    assert np.all(set_up_mesh.calculate_velocity(0, 1) == [8., 7.])
    assert np.all(set_up_mesh.calculate_velocity(0, 2) == [6., 11.])
    assert np.all(set_up_mesh.calculate_velocity(0, 3) == [6., 11.])


from dataclasses import dataclass
import itertools as iter
from collections import Counter
from typing import Tuple
from warnings import warn
import scipy.spatial as spspatial
import skimage as skimage
import collections
import numpy as np
import cv2
from PIL import Image

import forsys.vertex as vertex
import forsys.edge as edge
import forsys.cell as cell
import forsys.wkt as fwkt
import forsys.virtual_edges as fvedges

import forsys as fs

@dataclass
class Skeleton:
    """
    Reads and parses TIFF skeletons or CellPose npy files
    using OpenCV implementation to detect polygons

    :param fname: Path to segmented image
    :type fname: str
    :param mirror_y: False by default. Inverts the y-axis if true.
    :type mirror_y: bool, optional
    """
    fname: str
    mirror_y: bool = False
    minimum_distance: float = 0
    expand: int = 0

    def __post_init__(self):
        # TODO: Unify NPY and TIF, as they use similar OpenCV methods
        if self.fname.endswith(".tif") or self.fname.endswith(".tiff"):
            self.format = "tif"
            with Image.open(self.fname).convert("L") as curr_image_file:
                self.img_arr = np.array(curr_image_file)[1:-1, 1:-1]

            self.contours, _ = cv2.findContours(self.img_arr,
                                                cv2.RETR_TREE,
                                                # cv2.CHAIN_APPROX_SIMPLE,
                                                cv2.CHAIN_APPROX_NONE)
            self.contours = [np.vstack(p).squeeze() for p in self.contours][1:]
        elif self.fname.endswith(".npy"):
            self.format = "npy"
            np_data = np.load(self.fname,
                              allow_pickle=True).item()
            masks = np_data["masks"]
            border_id = -1
            if self.expand > 0:
                # add an additional mask of everything
                border_contour = skimage.measure.label(masks != 0)
                border_expand = skimage.segmentation.expand_labels(border_contour,
                                                                  distance=2)
                binary_mask_border = border_expand == 1
                polygon_contour, _ = cv2.findContours(binary_mask_border.astype(np.uint8),
                                                 cv2.RETR_TREE,
                                                 cv2.CHAIN_APPROX_NONE)
                which = np.argmax([c.shape[0] for c in polygon_contour])
                border_polygon = polygon_contour[which].astype(int).squeeze()

                border_id = np.max(masks) + 1
                for element in border_polygon:
                    masks[element[1], element[0]] = border_id

                masks = skimage.segmentation.expand_labels(masks,
                                                           distance=self.expand)

            self.contours = []
            for cell_n in np.unique(masks)[1:]:
                if cell_n == border_id:
                    continue
                binary_mask = masks == cell_n
                if binary_mask.sum() > 0:
                    current_contour, _ = cv2.findContours(binary_mask.astype(np.uint8),
                                                     cv2.RETR_TREE,
                                                     cv2.CHAIN_APPROX_NONE)

                # assert len(current_contour) == 1, f"More than one contour found for cell {cell_n}"
                which = np.argmax([c.shape[0] for c in current_contour])
                to_add = current_contour[which].astype(int).squeeze()
                self.contours.append(to_add)
        else:
            raise ValueError("File format not supported")

        # initialize the counters
        self.vertex_id = 0
        self.edge_id = 0
        self.cell_id = 0

        if self.mirror_y:
            warn("Legacy function, soon to be deprecated", DeprecationWarning, 2)
            self.max_y = float(max(list(iter.chain.from_iterable(
                [list(zip(*c))[1] for c in self.contours]))))

    def create_lattice(self, **kwargs) -> Tuple[dict, dict, dict]:
        """
        Creates vertices, edges and cells of the system in three dictioinaries

        :return: Vertices, edges and cells created from parsing the provided image
        :rtype: tuple
        """
        self.vertices = {}
        self.edges = {}
        self.cells = {}

        self.coords_to_key = {}

        self.edges_added = []

        rescale = kwargs.get("rescale", [1, 1])
        offset = kwargs.get("offset", [0, 0])

        if kwargs.get("max_cell_size", 0) != 0:
            all_cell_areas = [self.calculate_area(polygon)
                             for polygon in self.contours]
            ave_cell_area = np.mean(np.sort(all_cell_areas)[:-1])
            self.contours = [polygon for polygon in self.contours
                             if self.calculate_area(polygon) <
                             kwargs.get("max_cell_size") * ave_cell_area]

        for _, polygon in enumerate(self.contours):
            cell_vertices_list = []
            for coords in polygon:
                if self.mirror_y:
                    coords[1] = self.max_y - coords[1]
                vid = self.get_vertex_id_by_position(coords)

                if vid == -1:
                    vid = self.vertex_id
                    self.vertices[vid] = vertex.Vertex(int(vid),
                                                       (coords[0] - offset[0]) / rescale[0],
                                                       (coords[1] - offset[1]) / rescale[1])
                    self.vertex_id += 1

                    self.coords_to_key[tuple(coords)] = vid
                    # add vertex to this cell's list
                cell_vertices_list.append(self.vertices[vid])

            # join vertices with the edges
            number_of_vertices = len(polygon)
            for n in range(1, number_of_vertices):
                self.create_edge(n - 1, n, polygon)
            # Join last vertex with first
            self.create_edge(n, 0, polygon)

            # create cells now
            self.cells[self.cell_id] = cell.Cell(self.cell_id, cell_vertices_list, {})
            self.cell_id += 1

        if kwargs.get("reduce_amount", False):
            self.vertices, self.edges, self.cells = fwkt.reduce_amount(self.vertices,
                                                                    self.edges,
                                                                    self.cells)

        for current_cell in self.cells.values():
            # if any of the vertices in the cell has a vertex with only one cell, that is an external cell
            if np.any([len(v.ownCells) == 1 for v in current_cell.vertices]):
                current_cell.is_border = True
        for e in self.edges.values():
            # if at least one of the vertices have only one cell, this is an external edge
            if len(e.v1.ownCells) == 1 or len(e.v2.ownCells) == 1:
                e.external = True
            else:
                e.external = False

        self.main_vertices_map = {}
        if self.minimum_distance > 0:
            # TODO: It would be more efficient to use a dynamic KDtree
            all_points = np.array([[v.x, v.y] for v in self.vertices.values()])
            self.kdtree = spspatial.KDTree(all_points)
            vids_joined = []
            edges_to_delete = []
            for vid, vobj in self.vertices.items():
                # get all neighbors in the given distance, and join them to vid
                # in one vertex
                if vid not in vids_joined:
                    # it is a main vertex
                    neighbors = self.kdtree.query_ball_point([vobj.x, vobj.y],
                                                             self.minimum_distance)
                    self.main_vertices_map[vid] = None
                    joined_to_current_vid = []
                else:
                    continue

                for nn in neighbors:
                    # if the nn is is a main vertex, ignore
                    nn_vid = self.coords_to_key[tuple(all_points[nn])]
                    if nn_vid in self.main_vertices_map.keys() or nn_vid in vids_joined:
                        continue
                    else:
                        for cid in self.vertices[nn_vid].ownCells:
                            self.cells[cid].replace_vertex(self.vertices[nn_vid],
                                                           self.vertices[vid])
                        for eid in self.vertices[nn_vid].ownEdges:
                            self.edges[eid].replace_vertex(self.vertices[nn_vid],
                                                           self.vertices[vid])
                        vids_joined.append(nn_vid)
                        joined_to_current_vid.append(nn_vid)

                self.main_vertices_map[vid] = joined_to_current_vid

            for eid, eobj in self.edges.items():
                if (eobj.v1.id in vids_joined and eobj.v2.id in vids_joined) or eobj.v1.id == eobj.v2.id:
                    edges_to_delete.append(eid)

                elif bool(eobj.v1.id in vids_joined) ^ bool(eobj.v2.id in vids_joined):
                    if eobj.v1.id in vids_joined:
                        to_replace = eobj.v1.id
                    elif eobj.v2.id in vids_joined:
                        to_replace = eobj.v2.id
                    else:
                        raise ValueError("Inexistant vertex in the list of joined vertices")

                    for main_v, joined_to_it in self.main_vertices_map.items():
                        if to_replace in joined_to_it:
                            if (main_v in eobj.get_vertices_id()):
                                edges_to_delete.append(eid)
                            else:
                                eobj.replace_vertex(self.vertices[to_replace],
                                                    self.vertices[main_v])
                            break
            for ee in edges_to_delete:
                try:
                    del self.edges[ee]
                except KeyError:
                    print(f"**WARNING** edge {ee} not in list of edges")

            # check for repeated edges and same ones
            # TODO: This should be verified before, instead of repeating
            edges_to_delete = []
            # all_edges = [tuple(set(e.get_vertices_id())) for e in self.edges.values()]
            all_edges = [tuple(sorted(e.get_vertices_id())) for e in self.edges.values()]
            all_edges_counter = collections.Counter(all_edges)
            for current_edge, count in all_edges_counter.items():
                if count > 1:
                    edges_to_delete += [e.id for e in self.edges.values() if tuple(sorted(e.get_vertices_id())) == current_edge]
                    # keep one
                    edges_to_delete.pop()

            for ee in edges_to_delete:
                try:
                    del self.edges[ee]
                except KeyError:
                    print(f"**WARNING** edge {ee} not in list of edges")

            for nn in set(vids_joined):
                del self.vertices[nn]

            # list of all eddges indexing as tuples
            all_edges = {tuple(sorted(e.get_vertices_id())): e.id for e in self.edges.values()}
            non_edges = {tuple(sorted(e.get_vertices_id())): 0 for e in self.edges.values()}
            removed_vertices = []
            # find unused edges
            for cid, cobj in self.cells.items():
                cvertices = [c.id for c in cobj.vertices]
                for i_vid, vid in enumerate(cvertices):
                    if vid in removed_vertices:
                        continue
                    i_obj = self.vertices[vid]
                    next_vertex = cobj.get_next_vertex(i_obj)
                    previous_vertex = cobj.get_previous_vertex(i_obj)
                    if tuple(sorted([i_obj.id, next_vertex.id])) not in all_edges.keys():
                        intersection_own = self.get_cell_intersection(i_obj)
                        intersection_other = self.get_cell_intersection(next_vertex)

                        if cid in intersection_own and cid not in intersection_other:
                            # replace the vertex in the cell
                            cobj.replace_vertex(next_vertex, i_obj)
                            next_vertex.remove_cell(cid)
                            i_obj.add_cell(cid)
                            new_pair = tuple(sorted([i_obj.id, previous_vertex.id]))
                            removed_vertices.append(next_vertex.id)
                        elif cid not in intersection_own and cid in intersection_other:
                            cobj.replace_vertex(i_obj, next_vertex)
                            next_vertex.add_cell(cid)
                            i_obj.remove_cell(cid)
                            new_pair = tuple(sorted([next_vertex.id, previous_vertex.id]))
                            removed_vertices.append(i_obj.id)
                        else:
                            raise ValueError("No vertex in the cell intersection")
                        non_edges[new_pair] += 1
                    else:
                        non_edges[tuple(sorted([i_obj.id, next_vertex.id]))] += 1

            for n_eid, n_eid_count in non_edges.items():
                if n_eid_count == 0:
                    del self.edges[all_edges[n_eid]]

        self.all_big_edges = fvedges.create_edges_new(self.vertices,
                                                      self.cells)
        if self.format == "npy":
            # triangles in the middle
            self.triangular_holes()
        else:
            # triangles in the middle
            self.triangles_in_the_middle()

        self.remove_artifacts()

        return self.vertices, self.edges, self.cells

    def triangular_holes(self):
        vertices_of_triangles = {}
        for v in self.vertices.values():
            if len(v.ownEdges) >= 3:
                # this is could be a triangular hole
                # is it connected to two other TJ with only two cells ?
                cells_per_nn = self.is_vertex_in_hole(v)
                if len(cells_per_nn) > 0:
                    # if all are TJs, join in one point
                    vertices_of_triangles[v.id] = list(cells_per_nn.keys())
        # see if key is in at least 2 others
        appearances = {}
        for key, nns in vertices_of_triangles.items():
            appeard_in = []
            for key_2, nns2 in vertices_of_triangles.items():
                if key in nns2 and key != key_2:
                    appeard_in.append(key_2)
            appearances[key] = appeard_in
        triangles = []
        middle_edge = []
        for who, where in appearances.items():
            triangle = []
            if len(where) >= 2:
                triangle = sorted([who] + where)
                if triangle not in triangles:
                    triangles.append(triangle)
            elif len(where) == 0:
                # this is triangle in the middle of an edge
                middle_edge.append(who)
        # remove the one that are in the middle of the edge
        for middle in middle_edge:
            for ee in self.vertices[middle].ownEdges:
                vertices_ids = self.edges[ee].get_vertices_id()
                vertex_to_delete = vertices_ids[0] if vertices_ids[0] != middle else vertices_ids[1]
                if len(self.vertices[vertex_to_delete].ownCells) == 1:
                    # this is the one
                    break

            for cell_id in self.vertices[vertex_to_delete].ownCells:
                self.cells[cell_id].replace_vertex(self.vertices[vertex_to_delete],
                                                   self.vertices[middle])
            its_edges = list(self.vertices[vertex_to_delete].ownEdges)
            for edge_id in its_edges:
                del self.edges[edge_id]

            del self.vertices[vertex_to_delete]

        # Join the vertices in the triangles in the centroid
        for triangle in triangles:
            # get the centroid
            centroid = np.mean([self.vertices[v].get_coords() for v in triangle], axis=0)
            new_vid = self.get_new_vid()
            self.vertices[new_vid] = vertex.Vertex(int(new_vid),
                                                   centroid[0],
                                                   centroid[1])
            # replace the vertices with the new one
            for v in triangle:
                its_edges = list(self.vertices[v].ownEdges)
                for edge_id in its_edges:
                    if self.edges[edge_id].get_vertices_id()[0] in triangle and \
                        self.edges[edge_id].get_vertices_id()[1] in triangle:
                        del self.edges[edge_id]
                    else:
                        self.edges[edge_id].replace_vertex(self.vertices[v],
                                                           self.vertices[new_vid])
                for cell_id in self.vertices[v].ownCells:
                    self.cells[cell_id].replace_vertex(self.vertices[v],
                                                       self.vertices[new_vid])
                del self.vertices[v]

    def is_vertex_in_hole(self, v: fs.vertex.Vertex) -> dict:
        connected_edges_vertices = [self.edges[e].get_vertices_id() for e in v.ownEdges]
        connected_to = np.array(connected_edges_vertices).reshape(-1)
        connected_to = list(filter(lambda x: x != v.id, connected_to))
        # do at least two of them have only two cells?
        cells_per_nn = {vv: self.vertices[vv] for vv in connected_to}
        leq_2 = len(list(filter(lambda x: len(x.ownCells) <= 2, cells_per_nn.values())))
        eq_1 = len(list(filter(lambda x: len(x.ownCells) == 1, cells_per_nn.values())))
        if leq_2 >= 2 and eq_1 != 2 and len(v.ownCells) == 2:
            return cells_per_nn
        else:
            return {}

    def get_cell_intersection(self, vobj) -> list:
        connected_edges_vertices = [self.edges[e].get_vertices_id() for e in vobj.ownEdges]
        connected_to = np.unique(np.array(connected_edges_vertices).reshape(-1))
        cells_belonging = [set(self.vertices[v].ownCells) for v in connected_to]
        intersection = set.intersection(*cells_belonging)
        return list(intersection)

    def triangles_in_the_middle(self) -> None:
        # triangles in the middle
        inner_edge_triangles = []
        num_elements = len(self.all_big_edges)
        first_part = [(e[0], e[-1]) for e in self.all_big_edges]
        last_part = [(e[-1], e[0]) for e in self.all_big_edges]
        first_last = first_part + last_part
        inner_edge_triangles = [k for k, v in Counter(first_last).items() if v > 1]

        visited = []
        for index in range(len(inner_edge_triangles) - 1):
            if (inner_edge_triangles[index] in visited or
                inner_edge_triangles[index][::-1] in visited):
                continue

            edge_0_index = first_last.index(inner_edge_triangles[index]) % num_elements
            edge_1_index = first_last.index(inner_edge_triangles[(index + 1)]) % num_elements
            edge_0 = self.all_big_edges[edge_0_index]
            if len(edge_0) > 3:
                continue
            edge_1 = self.all_big_edges[edge_1_index]
            vertex_id_to_delete = list(set(edge_0).symmetric_difference(set(edge_1)))[0]
            # vertex_id_to_delete = np.intersect1d(edge_0, edge_1)[0]

            its_cells = self.vertices[vertex_id_to_delete].ownCells
            for cell_id in its_cells:
                self.cells[cell_id].replace_vertex(self.vertices[vertex_id_to_delete],
                                                   self.vertices[edge_0[0]])

            its_edges = list(self.vertices[vertex_id_to_delete].ownEdges)
            for edge_id in its_edges:
                del self.edges[edge_id]

            del self.vertices[vertex_id_to_delete]
            # visited.append(inner_edge_triangles[index])
    #     print(f"Deleted {len(visited)} triangles in the middle of the system")

    def remove_artifacts(self) -> None:
        # get artifacts from the contour and apply T3 transitions to them
        all_artifact_vertices = list(self.get_artifacts())

        artifacts = []
        while len(all_artifact_vertices) != 0:
            current_artifact = [all_artifact_vertices[0]]
            current_len0 = -1
            while current_len0 != len(current_artifact):
                current_artifact = self.add_vertices_to_current(all_artifact_vertices, current_artifact)
                current_len0 = len(current_artifact)
            # remove vertices already in artifact from all_artifact_vertices
            for ii in current_artifact:
                all_artifact_vertices.remove(ii)

            artifacts.append(current_artifact)

        # now make a T3 transition for each triangle:
        for artifact in artifacts:
            self.do_t3_transition(artifact)

        # if all vertices in a cell only have that cell as own, delete cell
        cells_to_remove = []
        for c in self.cells.values():
            vertices_le_one = [len(v.ownCells) <= 1 for v in c.vertices]

            if np.all(vertices_le_one):
                cells_to_remove.append(c.id)
                for v in c.vertices:
                    # remove edges
                    for e in v.ownEdges:
                        del self.edges[e]
                    # remove vertices
                    try:
                        del self.vertices[v.id]
                    except KeyError:
                        print(f"**WARNING** vertex {v.id} not in list of vertices")
        # remove the cell
        for cid in cells_to_remove:
            del self.cells[cid]

    def add_vertices_to_current(self, artifact_vertices: list, current_artifact: list) -> list:
        """
        Creates the list of vertices of a given artifact
        :param artifact_vertices: All vertices belonging to artifacts
        :type artifact_vertices: list
        :param current_artifact: List to append the vertices of the current artifact
        :type current_artifact: list

        :return: Vertices, edges and cells created from parsing the provided image
        :rtype: list
        """
        v0 = current_artifact[-1]
        for eid in self.vertices[v0].ownEdges:
            other_vertex = self.edges[eid].get_other_vertex_id(v0)
            if other_vertex in artifact_vertices and other_vertex not in current_artifact:
                current_artifact.append(other_vertex)
        return current_artifact

    def do_t3_transition(self, artifact: list):
        """
        Perform a T3 transition in all the identified artifacts

        :param artifact: List with all the artifacts detected in the system
        :type artifact: list
        :return: None
        """
        # get the mid point and create a vertex there
        ave_x = np.mean([self.vertices[ii].x for ii in artifact])
        ave_y = np.mean([self.vertices[ii].y for ii in artifact])

        new_vid = self.get_new_vid()
        self.vertices[new_vid] = vertex.Vertex(int(new_vid),
                                               ave_x,
                                               ave_y)

        # get the edge and join them to the new point)
        for v in artifact:
            edge_to_remove = []
            edge_to_replace = []
            for e in self.vertices[v].ownEdges:
                if (self.edges[e].v1.id in artifact and self.edges[e].v2.id in artifact) or \
                        (self.edges[e].v2.id in artifact and self.edges[e].v1.id in artifact):
                    edge_to_remove.append(e)
                else:
                    edge_to_replace.append(e)

            for ii in range(0, len(edge_to_remove)):
                del self.edges[edge_to_remove[ii]]

            for ii in range(0, len(edge_to_replace)):
                edge_object = self.edges[edge_to_replace[ii]]
                # if new_vid != edge_object.v1.id and new_vid != edge_object.v2.id:
                edge_object.replace_vertex(self.vertices[v], self.vertices[new_vid])

            # add the cells corresponding
            for c in self.vertices[v].ownCells:
                # cell_vertex_ids = [v.id for v in self.cells[c]]
                self.cells[c].replace_vertex(self.vertices[v], self.vertices[new_vid])

        for jj in range(0, len(artifact)):
            if len(self.vertices[artifact[jj]].ownEdges) == 0:
                del self.vertices[artifact[jj]]
            else:
                print("WARNING: BAD VERTEX DELETION IN TRIANGLE - MORE THAN ZERO EDGE")

    def get_artifacts(self) -> list:
        """
        Create list with all the artifacts detected in the processing of the tissue

        :return: List of vertices belonging to artifacts
        :rtype: list
        """
        # triangles with TJs
        candidate_artifact_vertices = []
        for v in self.vertices.values():
            if len(v.ownEdges) > len(v.ownCells):
                # this is part of a triangle
                candidate_artifact_vertices.append(v.id)

        # list of external vertices:
        external_vertices = []
        for e in self.edges.values():
            if e.external:
                external_vertices.append(e.v1.id)
                external_vertices.append(e.v2.id)
        artifact_vertices = np.setdiff1d(candidate_artifact_vertices, external_vertices)

        return artifact_vertices

    def create_edge(self, n0: int , n1: int, polygon: list):
        """
        Add tuples of vertices identified by the vertices that from them and create the SmallEdge object

        :param n0: ID in the polygon list of the first vertex
        :type n0: int
        :param n1: ID in the polygon list of the second vertex
        :type n1: int
        :param n0: List of vertices of the current polygon
        :type polygon: int
        """
        # TODO: Functionalize function.
        v1 = self.vertices[self.get_vertex_id_by_position(polygon[n0])]
        v2 = self.vertices[self.get_vertex_id_by_position(polygon[n1])]
        if not (v1.id, v2.id) in self.edges_added and not (v2.id, v1.id) in self.edges_added:
            self.edges[self.edge_id] = edge.SmallEdge(self.edge_id, v1, v2)
            self.edge_id += 1
            self.edges_added.append((v1.id, v2.id))

    def get_vertex_id_by_position(self, coordinates: list) -> int:
        """
        Check whether vertices were already created using a hashmap

        :param coordinates: tuple with the coordiantes of the vertex to check
        :type coordinates: list

        :return: Current key of the given coordinates or -1 if there are no vertices at that position
        :rype: int
        """
        try:
            return self.coords_to_key[tuple(coordinates)]
        except KeyError:
            return -1


    def get_new_vid(self) -> int:
        """
        Returns the best possible ID for the creation of a new vertex

        :return: unused ID in the vertices dictionary
        :rtype: int
        """
        new_vid = max(self.vertices.keys()) + 1
        while new_vid in self.vertices.keys():
            new_vid += 1
        return new_vid

    @staticmethod
    def calculate_area(vertices: list) -> float:
        """
        Calculate the area of a given polygon using the shoelace formula

        :param vertices: array of [x,y] positions of the ordered vertices in a polygon
        :type vertices: list
        :return: Area of the polygon described by the vertices
        :rtype: float
        """
        x = [i[0] for i in vertices]
        y = [i[1] for i in vertices]
        return abs(0.5 * (np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1))))

    @staticmethod
    def get_distance(v1: vertex.Vertex, v2: vertex.Vertex) -> float:
        """
        Get the distance between two vertices in the system

        :param v1: First vertex
        :type v1: vertex.Vertex
        :param v2: Second vertex
        :type v2: vertex.Vertex
        :return: Distance between the vertices
        :rtype: float
        """
        return np.linalg.norm(np.array([v1.x, v1.y]) - np.array([v2.x, v2.y]))



import numpy as np
from dataclasses import dataclass
import scipy.optimize as scop
from typing import Tuple, Optional


@dataclass
class GeneralMatrix:
    """Top level class to generalize solving the matrix system. Serves as parent for 
    pressure and stress matrix

    :param frame: Frame object from which matrix system is created
    :type frame: object
    :param timeseries: Time series dictionary of frames
    :type timeseries: dict
    """
    frame: object
    timeseries: dict

    def __post_init__(self):
        """Constructor method
        """
        self.map_vid_to_row = {}
        
        self.externals_to_use = []
        self.big_edges_to_use = self.frame.internal_big_edges
        self.mapping_order = {}

        self.lhs_matrix = None
        self.rhs_matrix = None


    def solve_system(self, **kwargs) -> list:
        """Solve the system of equations. The LHS and RHS have to be set by the child classes beforehand.

        :return: Returns a list with the values for the solution.
        :rtype: list
        """
        np.seterr(all='raise')
        assert self.lhs_matrix is not None, "LHS matrix not set"
        assert self.rhs_matrix is not None, "RHS matrix not set"

        lhs_matrix_ls = self.lhs_matrix.T @ self.lhs_matrix
        rhs_matrix_ls = self.lhs_matrix.T @ self.rhs_matrix  # vector

        removed_index = None
        if kwargs.get("method") == "fix_stress":
            lhs_matrix_ls, rhs_matrix_ls, removed_index = self.fix_one_stress(lhs_matrix_ls, rhs_matrix_ls)
        elif kwargs.get("method") == "lagrange_pressure":
            lhs_matrix_ls, rhs_matrix_ls = self.add_lagrange_multiplier(lhs_matrix_ls, rhs_matrix_ls, 0.)
        
        lhs_matrix_ls = lhs_matrix_ls.astype(np.float64)
        rhs_matrix_ls = rhs_matrix_ls.astype(np.float64)

        try:
            xres = np.linalg.inv(lhs_matrix_ls) @ rhs_matrix_ls

            if np.any(xres[:-1] < 0) and not kwargs.get("allow_negatives", True):
                print("Numerically solving due to negative values")
                xres, _ = scop.nnls(lhs_matrix_ls, rhs_matrix_ls, maxiter=kwargs.get("nnls_max_iter"))

        except np.linalg.LinAlgError:
            print("Numerically solving due to singular matrix")
            xres, _ = scop.nnls(lhs_matrix_ls, rhs_matrix_ls, maxiter=kwargs.get("nnls_max_iter"))

        if removed_index is not None:
            xres = np.insert(xres, removed_index, 1.)
        
        self.solution = xres[:-1].tolist()
        for val in self.mapping_order.values():
            if val in self.removed_columns:
                self.solution.insert(val, 0.)

        return self.solution

    def add_lagrange_multiplier(self,
                                lhs_matrix_ls: np.ndarray,
                                rhs_matrix_ls: np.ndarray,
                                constraint: float = 0.
                                ) -> Tuple:
        """Add a constraint to the matrix system through a Lagrange multiplier. 
        LHS and RHS matrices should be already in least-squares form.

        :param lhs_matrix_ls: LHS matrix in least-squares form
        :type lhs_matrix_ls: np.ndarray
        :param rhs_matrix_ls: RHS matrix in least-squares form
        :type lhs_matrix_ls: np.ndarray
        :param constraint: Value to give as constraint
        :type constraint: float
        :return: LHS and RHS matrices with the additional Lagrange multiplier.
        :rtype: Tuple
        """
        lhs_rows, lhs_cols = lhs_matrix_ls.shape
        lhs_matrix_ls = np.vstack((lhs_matrix_ls, np.ones(lhs_cols)))

        # Now insert the two corresponding cols for the lagrange multipliers
        cMatrix = np.array([1.] * lhs_rows + [0.])
        lhs_matrix_ls = np.hstack((lhs_matrix_ls, cMatrix.reshape(-1, 1)))

        rhs_matrix_ls = np.concatenate((rhs_matrix_ls, [constraint]))

        return lhs_matrix_ls, rhs_matrix_ls

    @staticmethod
    def fix_one_stress(lhs_matrix_ls: np.ndarray,
                       rhs_matrix_ls: np.ndarray,
                       column_to_fix: Optional[int] = None,
                       value_to_fix_to: float = 1.
                       ) -> Tuple:
        """Fix one of the values in the solution. This could be used to give a scale to the solution.

        :param lhs_matrix_ls: LHS matrix in least-squares form
        :type lhs_matrix_ls: np.ndarray
        :param rhs_matrix_ls: RHS matrix in least-squares form
        :type rhs_matrix_ls: np.ndarray
        :param column_to_fix: ID of unknown to fix, defaults to None
        :type column_to_fix: int, optional
        :param value_to_fix_to: Value to give the fixed unknown, defaults to 1
        :type value_to_fix_to: float, optional
        :return: Returns the values of the LHS and RHS matrix and the new number of unknowns
        :rtype: Tuple
        """
        # Replace column with the most connections
        if column_to_fix is None:
            non_zero_count = np.count_nonzero(lhs_matrix_ls, axis=0)
            max_index = np.argmax(non_zero_count)
            rhs_matrix_ls = rhs_matrix_ls - value_to_fix_to * lhs_matrix_ls[:, max_index]
            lhs_matrix_ls = np.delete(lhs_matrix_ls, max_index, 1)
        else:
            raise NotImplementedError(f'{column_to_fix=}')

        return lhs_matrix_ls, rhs_matrix_ls, max_index



import pytest

import os
import numpy as np
import forsys as fs

@pytest.fixture
def set_up_cell():
    def get_hexagon_vertices(apothem=1):
        hex_vertices = [[apothem * np.cos(ang), apothem * np.sin(ang)] 
                        for ang in np.arange(0, 2*np.pi, 60 * np.pi / 180)]
        return hex_vertices

    vertices = {}
    edges = {}

    for vid, vertex in enumerate(get_hexagon_vertices()):
        vertices[vid] = fs.vertex.Vertex(vid, vertex[0], vertex[1])
        if vid > 0:
            edges[vid - 1] = fs.edge.SmallEdge(vid - 1, vertices[vid - 1], vertices[vid])
    edges[len(edges)] = fs.edge.SmallEdge(len(edges), vertices[0], vertices[vid])


    cell = fs.cell.Cell(0, list(vertices.values()), {})


    yield cell

@pytest.fixture()
def set_up_4_cells():
    filename = os.path.join("tests", "data", "skeleton", "triangle_holes.npy")
    skeleton = fs.skeleton.Skeleton(filename,
                                    minimum_distance=7)
    vertices, edges, cells = skeleton.create_lattice()
    frame = fs.frames.Frame(0,
                            vertices,
                            edges,
                            cells)
    yield frame

def test_get_area(set_up_cell):
    assert abs(np.round(set_up_cell.get_area(), 6)) == np.round(1.5 * np.sqrt(3), 6)


def test_get_area_sign(set_up_cell):
    assert set_up_cell.get_area_sign() == -1
    set_up_cell.vertices = set_up_cell.vertices[::-1]
    assert set_up_cell.get_area_sign() == 1


def test_get_perimeter(set_up_cell):
    assert int(set_up_cell.get_perimeter()) == 6


def test_get_edges(set_up_4_cells):
    # TODO: Create a cell from scratch to make it independent of reading npy
    edges_cell_1 = [542, 545, 552, 561, 570, 578, 586, 594, 602, 611, 620, 628,
                    1678, 1670, 1661, 1654, 1645, 1637, 1629, 1620, 1612, 1603,
                    1595, 1586, 1576, 1568, 1804, 1797, 1789, 1781, 1773, 2337,
                    2330, 2322, 2314, 2306, 2297, 2290, 2282, 2273, 2264, 896,
                    904, 912, 921, 929, 937, 946, 955, 963, 971, 980, 988, 996,
                    1004, 1012, 1021, 1030, 1038, 1047, 1055, 1064, 1072, 1080,
                    1089, 1097, 1106, 1114, 1123, 1132, 1141, 1150, 1158, 1167]
    assert set_up_4_cells.cells[1].get_edges() == edges_cell_1



import pytest
import numpy as np
import pandas as pd
import copy
import os
import forsys as fs

@pytest.fixture
def set_up_frame():
    vertices = {}
    edges = {}
    cells = {}
    frame = fs.frames.Frame(0, vertices, edges, cells)


    for vid in range(10):
        vertices[vid] = fs.vertex.Vertex(vid, vid + 1 , vid + 1)
        if vid > 0:
            edges[vid - 1] = fs.edge.SmallEdge(vid - 1, vertices[vid - 1], vertices[vid])

    frame.big_edges = {
        0: fs.edge.BigEdge(0,
                   vertices=[vertices[0], vertices[1], vertices[2], vertices[3], vertices[4]]),
        1: fs.edge.BigEdge(1, 
                   vertices=[vertices[5],
                             vertices[6],
                             vertices[7],
                             vertices[8],
                             vertices[9]])
    }
    return frame


@pytest.fixture
def furrow():
    frames = {}
    for ii in range(0, 2):
        surfaceEvolver = fs.surface_evolver.SurfaceEvolver(os.path.join(
                                                            "tests", 
                                                            "data", 
                                                            "furrow_gauss_velocity", 
                                                            f"stage{ii}.dmp"))
        frames[ii] = fs.frames.Frame(ii,
                                    surfaceEvolver.vertices,
                                    surfaceEvolver.edges, 
                                    surfaceEvolver.cells, 
                                    time=ii, 
                                    gt=True)
    forsys = fs.ForSys(frames, cm=False)
    yield forsys

@pytest.fixture
def set_up_frame_with_small_edge():
    vertices = {}
    edges = {}
    cells = {}
    frame = fs.frames.Frame(0, vertices, edges, cells)

    vertices[0] = fs.vertex.Vertex(0, 0, 1)
    vertices[1] = fs.vertex.Vertex(1, 1, 2)
    vertices[2] = fs.vertex.Vertex(2, 2, 3)

    edges[0] = fs.edge.SmallEdge(0, vertices[0], vertices[1])
    edges[1] = fs.edge.SmallEdge(1, vertices[1], vertices[2])


    frame.big_edges = {
        0: fs.edge.BigEdge(0,
                    vertices=[vertices[0], vertices[1]])
    }

    return frame


def test_filter_edges_sg(set_up_frame):
    set_up_frame.filter_edges(method="SG")
    # Check if Savitzky-Golay filter is applied correctly
    assert np.allclose(set_up_frame.big_edges[0].xs, [1.0, 2.0, 3.0, 4.0, 5.0])
    assert np.allclose(set_up_frame.big_edges[0].ys, [1.0, 2.0, 3.0, 4.0, 5.0])
    assert np.allclose(set_up_frame.big_edges[1].xs, [6.0, 7.0, 8.0, 9.0, 10.0])
    assert np.allclose(set_up_frame.big_edges[1].ys, [6.0, 7.0, 8.0, 9.0, 10.0])


def test_filter_edges_none(set_up_frame):
    set_up_frame.filter_edges(method="none")
    # Check if no filtering is applied
    assert np.allclose(set_up_frame.big_edges[0].xs, [1.0, 2.0, 3.0, 4.0, 5.0])
    assert np.allclose(set_up_frame.big_edges[0].ys, [1.0, 2.0, 3.0, 4.0, 5.0])
    assert np.allclose(set_up_frame.big_edges[1].xs, [6.0, 7.0, 8.0, 9.0, 10.0])
    assert np.allclose(set_up_frame.big_edges[1].ys, [6.0, 7.0, 8.0, 9.0, 10.0])


def test_filter_edges_invalid_method(set_up_frame):
    with pytest.raises(ValueError):
        set_up_frame.filter_edges(method="invalid")


def test_filter_edges_invalid_frame(set_up_frame_with_small_edge):
    initial_big_edges = copy.deepcopy(set_up_frame_with_small_edge.big_edges)
    set_up_frame_with_small_edge.filter_edges(method="SG")
    assert set_up_frame_with_small_edge.big_edges == initial_big_edges


def test_get_cell_properties_df(furrow):
    furrow.build_force_matrix(when=0)
    furrow.solve_stress(when=0)

    furrow.build_pressure_matrix(when=0)
    furrow.solve_pressure(when=0, method="lagrange_pressure")

    df = furrow.frames[0].get_cell_properties_df(center_method="centroid")
    assert isinstance(df, pd.DataFrame)
    assert "cid" in df.columns and df["cid"].dtype == int
    assert "centerx" in df.columns and df["centerx"].dtype == float
    assert "centery" in df.columns  and df["centery"].dtype == float
    assert "perimeter" in df.columns and df["perimeter"].dtype == float
    assert "area" in df.columns and df["area"].dtype == float
    assert "pressure" in df.columns and df["pressure"].dtype == float
    assert len(df) == len(furrow.frames[0].cells)

def test_get_edges_props_df(furrow):
    furrow.build_force_matrix(when=0)
    furrow.solve_stress(when=0)

    df = furrow.frames[0].get_edges_props_df()
    assert isinstance(df, pd.DataFrame)
    assert "eid" in df.columns and df["eid"].dtype == int
    assert "posx" in df.columns and df["posx"].dtype == float
    assert "posy" in df.columns  and df["posy"].dtype == float
    assert "tension" in df.columns and df["tension"].dtype == float
    assert "gt_tension" in df.columns and df["gt_tension"].dtype == float
    assert len(df) == len(furrow.frames[0].big_edges)


from PIL import Image
import numpy as np
import itertools
import math
from scipy import interpolate

def read_myosin(frame,
                tiff_path,
                integrate=False,
                normalize="average",
                layers: int = 1,
                **kwargs):
    image = Image.open(tiff_path)

    if kwargs.get("use_all", False):
        big_edges_to_use = frame.big_edges_list
    else:
        big_edges_to_use = frame.internal_big_edges

    return get_intensities(big_edges_to_use,
                           image,
                           integrate,
                           normalize,
                           layers,
                           **kwargs)


def get_intensities(big_edges,
                    image,
                    integrate=False,
                    normalize="average",
                    layers: int = 1,
                    **kwargs):
    intensities = {}

    image_to_use = image

    for be_id, big_edge in enumerate(big_edges):
        intensities_per_edge = []
        if integrate:
            vertices, length = get_interpolation(big_edge, layers, **kwargs)
            intensity_to_use = sum(map(image_to_use.getpixel,
                                       vertices)) / length
        else:
            for vertex in big_edge.vertices:
                intensity = get_intensity(image_to_use, vertex, layers, **kwargs)
                intensities_per_edge.append(intensity)

            intensity_to_use = np.mean(list(map(np.median,
                                                intensities_per_edge)))
        try:
            key_to_use = big_edges.index(big_edge)
        except ValueError:
            key_to_use = "ext_"+str(be_id)

        intensities[key_to_use] = intensity_to_use
        # big_edge.gt = intensities[key_to_use]

    intensities_only_internal = {k: v for k, v in intensities.items() if not isinstance(k, str)}

    if normalize == "maximum":
        raise NotImplementedError
    elif normalize == "average":
        mean_value = np.mean(list(intensities_only_internal.values()))
        intensities_only_internal = {k: v/mean_value for k, v in intensities_only_internal.items()}

    
    for be_id, big_edge in enumerate(big_edges):
        big_edge.gt = intensities_only_internal[be_id]

    return intensities_only_internal


def get_intensity(image, vertex, layers=1, **kwargs):
    offset = kwargs.get("offset", [0, 0])
    rescale = kwargs.get("rescale", [1, 1])
    x_y_position = [(vertex.x * rescale[0]) + offset[0], (vertex.y * rescale[1]) + offset[1]]
    pixel_positions = get_layer_elements(x_y_position, layers)
    return list(map(image.getpixel, pixel_positions))


def get_layer_elements(position, layers=1):
    pixel_positions = []
    layer_range = np.arange(-layers, layers + 1)
    for (ii, kk) in itertools.product(layer_range, layer_range):
        xy_pixel = (position[0] + ii, position[1] + kk)
        pixel_positions.append(xy_pixel)
    return pixel_positions


def get_interpolation(big_edge, layers, **kwargs):
    offset = kwargs.get("offset", [0, 0])
    rescale = kwargs.get("rescale", [1, 1])
    length = 0
    all_vertices = set()
    xs_values = [(value * rescale[0]) + offset[0] for value in big_edge.xs]
    ys_values = [(value * rescale[1]) + offset[1] for value in big_edge.ys]
    # xy_pairs = list(zip(big_edge.xs, big_edge.ys))
    xy_pairs = list(zip(xs_values, ys_values))

    for ii in range(1, len(xy_pairs)):
        current_vertex = list(map(math.ceil, xy_pairs[ii - 1]))
        next_vertex = list(map(math.ceil, xy_pairs[ii]))
        all_vertices.update(walk_two_vertices(current_vertex, next_vertex, layers))
        length += np.linalg.norm(np.array(xy_pairs[ii - 1]) - np.array(xy_pairs[ii]))

    return all_vertices, length


def walk_two_vertices(v0, v1, layers):
    vertices_to_return = set()
    diff_x = abs(v0[0] - v1[0])
    diff_y = abs(v0[1] - v1[1])
    axis = 0 if diff_x > diff_y else 1
    delta = 1 if v0[axis] < v1[axis] else -1
    interpolation = interpolate.interp1d([v0[axis], v1[axis]],
                                                [v0[axis - 1], v1[axis - 1]],
                                                kind="linear")
    for value in range(v0[axis], v1[axis], delta):
        if axis == 0:
            position = (value, interpolation(value))
        else:
            position = (interpolation(value), value)

        vertices_to_return.update(get_layer_elements(position, layers))
    return vertices_to_return


import pytest
import forsys as fs
import numpy as np
from PIL import Image

from scipy import interpolate

line_types = ["horizontal", "vertical", "straight_line", "curved", "general"]
layers = [0, 1, 2]

truth_values_average = {
    "horizontal-0": 100.0,
    "vertical-0": 100.0,
    "straight_line-0": 100.0,
    "curved-0": 100.0,
    "general-0": 100.0
}

def generate_xy_values(line_type, n_vertices, pixel_spacing, x_offset):
    regular_spacing = [x_offset + pixel_spacing * ii for ii in range(n_vertices)]

    if line_type == "horizontal":
        x_values = regular_spacing
        y_values = np.repeat(5, n_vertices)
    elif line_type== "vertical":
        x_values = np.repeat(5, n_vertices)
        y_values = regular_spacing
    elif line_type == "straight_line":
        x_values = regular_spacing
        y_values = regular_spacing
    elif line_type == "general":
        x_values = regular_spacing
        y_values = [int(ii + np.random.uniform(-5, 5)) for ii in x_values]
    elif line_type == "curved":
        x_values = regular_spacing
        y_values = [int(np.sqrt(40**2 - (ii - x_offset)**2)) for ii in regular_spacing]
    else:
        raise NotImplementedError
    
    return x_values, y_values


@pytest.fixture(params=line_types)
def set_up_big_edges(request):
    n_vertices = 20

    x_values, y_values = generate_xy_values(request.param, n_vertices, 2, 20)
    
    vertices = {}
    edges = {}
    truth_values = {"integration": {0: 0,
                                    1: 0,
                                    2: 0}}
    
    image_data = np.zeros((200, 200))
        
    vid = 0

    truth_ave_l0 = []
    truth_ave_l1 = []
    truth_ave_l2 = []

    for vid in range(n_vertices):
        vertices[vid] = fs.vertex.Vertex(vid, x_values[vid], y_values[vid])
        image_data[y_values[vid], x_values[vid]] = 100
        if vid > 0:
            edges[vid - 1] = fs.edge.SmallEdge(vid - 1, vertices[vid - 1], vertices[vid])

        elements_plus_one = fs.myosin.get_layer_elements((x_values[vid], y_values[vid]), 1)
        elements_plus_two = fs.myosin.get_layer_elements((x_values[vid], y_values[vid]), 2)
        for x_val, y_val in elements_plus_one:
            if image_data[y_val, x_val] != 100:
                image_data[y_val, x_val] = 50
        for x_val, y_val in elements_plus_two:
            if image_data[y_val, x_val] != 100 and image_data[y_val, x_val] != 50:
                image_data[y_val, x_val] = 20

    big_edge = fs.edge.BigEdge(0, list(vertices.values()))
    image = Image.fromarray(image_data)
    
    # quantification
    for vid in range(n_vertices):
        per_vertex_layers = {0: [],
                             1: [],
                             2: []}
        # average
        for layer in layers:
            elements_in_the_layer = fs.myosin.get_layer_elements((x_values[vid], y_values[vid]), layer)
            for x_val, y_val in elements_in_the_layer: 
                per_vertex_layers[layer].append(image_data[y_val, x_val])

                # n_elements[layer] += 1

        truth_ave_l0.append(np.median(per_vertex_layers[0]))
        truth_ave_l1.append(np.median(per_vertex_layers[1]))
        truth_ave_l2.append(np.median(per_vertex_layers[2]))

    # integration
    interpolation = lambda x: int(interpolate.interp1d(big_edge.xs, big_edge.ys, kind="linear")(x))
    xrange = range(min(big_edge.xs), max(big_edge.xs) + 1)

    elements_0 = {element for x_val in xrange for element in fs.myosin.get_layer_elements((x_val, interpolation(x_val)), 0)}
    elements_1 = {element for x_val in xrange for element in fs.myosin.get_layer_elements((x_val, interpolation(x_val)), 1)}
    elements_2 = {element for x_val in xrange for element in fs.myosin.get_layer_elements((x_val, interpolation(x_val)), 2)}

    truth_values["integration"][0] = sum([image_data[int(y_val), int(x_val)] for x_val, y_val in elements_0]) / len(elements_0)
    truth_values["integration"][1] = sum([image_data[int(y_val), int(x_val)] for x_val, y_val in elements_1]) / len(elements_1)
    truth_values["integration"][2] = sum([image_data[int(y_val), int(x_val)] for x_val, y_val in elements_2]) / len(elements_2)

    truth_values["average"] = {
                                0: np.mean(truth_ave_l0),
                                1: np.mean(truth_ave_l1),
                                2: np.mean(truth_ave_l2)
                                }

    yield [big_edge], image, truth_values


@pytest.mark.parametrize("layers", layers)
# new myosin integration makes test obsolte. Should rewrite test
# @pytest.mark.parametrize("integrating", [True, False])
@pytest.mark.parametrize("integrating", [False])
def test_get_intensities(set_up_big_edges, integrating, layers):
    big_edge, image, truth_values = set_up_big_edges

    myosin = fs.myosin.get_intensities(big_edge,
                              image,
                              integrate=integrating,
                              normalize=None,
                              corrected_by_tilt=False,
                              layers=layers)
    
    calculation = "integration" if integrating else "average"
    assert round(myosin[0], 3) == round(truth_values[calculation][layers], 3)


@pytest.mark.parametrize("layers", layers)
def test_get_layer_elements(layers):
    elements = fs.myosin.get_layer_elements([0, 0], layers)
    layer1 = set([(0, 0),
                  (-1, 0),
                  (-1, 1),
                  (0, 1),
                  (1, 1),
                  (1, 0),
                  (1, -1),
                  (0, -1),
                  (-1, -1),
                  (-1, 0)])
    layer2 = set([(-2, 0),
                  (-2, 1),
                  (-2, 2),
                  (-1, 2),
                  (0, 2),
                  (1, 2),
                  (2, 2),
                  (2, 1),
                  (2, 0),
                  (2, -1),
                  (2, -2),
                  (1, -2),
                  (0, -2),
                  (-1, -2),
                  (-2, -2),
                (-2, -1)])
    layer2.update(layer1)
    if layers == 0:
        assert elements == [(0, 0)]
    elif layers == 1:
        assert set(elements) == set(layer1)
    elif layers == 2:
        assert set(elements) == set(layer2)



import os
import json

import forsys as fs
import matplotlib.pyplot as plt


def create_folders(folder_to_create):
    if not os.path.exists(folder_to_create):
        os.makedirs(os.path.join(folder_to_create, "connections"))
        os.makedirs(os.path.join(folder_to_create, "myosin"))
        for folder_name in ["static", "dynamic"]:
            os.makedirs(os.path.join(folder_to_create,
                                     folder_name,
                                     "csvs"))
            os.makedirs(os.path.join(folder_to_create,
                                     folder_name,
                                     "fit_per_time"))
            os.makedirs(os.path.join(folder_to_create,
                                     folder_name,
                                     "tissues"))
            os.makedirs(os.path.join(folder_to_create,
                                     folder_name,
                                     "forces"))
            os.makedirs(os.path.join(folder_to_create,
                                     folder_name,
                                     "pressures"))
            os.makedirs(os.path.join(folder_to_create,
                                     folder_name,
                                     "stress_tensor"))
            
def create_folders_sweep(folder_to_create):
    if not os.path.exists(folder_to_create):
        os.makedirs(os.path.join(folder_to_create, "connections"))
        os.makedirs(os.path.join(folder_to_create, "myosin"))
        os.makedirs(os.path.join(folder_to_create, "fit_per_time"))
        os.makedirs(os.path.join(folder_to_create, "csvs"))

def create_directory(name, upperFolder):
    directory = os.path.join(upperFolder, name)
    if not os.path.exists(directory):
        os.makedirs(directory)

def load_initial_guess(guess_file, min_time, max_time):
    try:
        with open(guess_file) as jfile:
            initial_guess = json.load(jfile)
            initial_guess = {int(k): {int(kin): vin for kin, vin in v.items()} for k, v in initial_guess.items()}
    except (FileNotFoundError, TypeError):
        initial_guess = {}
        print("No guess file, using zero guess")
    number_of_frames = max_time - min_time

    initial_guess = {k: {} for k in range(number_of_frames)
                     if k not in initial_guess.keys()} | initial_guess
    return initial_guess


def create_plots(frame_number, forsys, res_folder, myo=False, pressure=True, compress=1):
    vertices = forsys.frames[frame_number].vertices
    edges = forsys.frames[frame_number].edges
    cells = forsys.frames[frame_number].cells
    # mesh
    fs.plot.plot_mesh(vertices,
                        edges,
                        cells,
                        f"mesh_{frame_number}.png",
                        f"{res_folder}/tissues/",
                        mirror_y=True)
    # stresses
    fs.plot.plot_inference(forsys.frames[frame_number],
                            step=frame_number,
                            folder=os.path.join(res_folder, "forces"),
                            normalized="absolute",
                            mirror_y=False,
                            colorbar=False,
                            compress_scale=compress)
    print("Saving to ", os.path.join(res_folder, "forces", f"{frame_number}.png"))
    plt.savefig(os.path.join(res_folder, "forces", f"{frame_number}.png"), dpi=350)
    plt.close()
    if myo:
        # myosin
        print("Plotting myosin")
        fs.plot.plot_inference(forsys.frames[frame_number],
                               ground_truth=True,
                                step=frame_number,
                                folder=os.path.join(res_folder, "forces"),
                                normalized="absolute",
                                mirror_y=False,
                                colorbar=False)
        plt.savefig(os.path.join(res_folder, "myosin", f"{frame_number}.png"), dpi=350)
        plt.close()
    
    if pressure:
        fs.plot.plot_inference(forsys.frames[frame_number],
                                step=frame_number,
                                pressure=True,
                                folder=os.path.join(res_folder, "pressures"),
                                normalized="max",
                                mirror_y=False,
                                colorbar=False)
        plt.savefig(os.path.join(res_folder, "pressures", f"{frame_number}.png"), dpi=350)
        plt.close()
    
        fs.plot.plot_stress_tensor(forsys.frames[frame_number], 
                        os.path.join(res_folder, "stress_tensor"), 
                        frame_number, 
                        grid=12, 
                        radius=5,
                        tensor_scale=1.5)




if __name__ == '__main__':
    import importlib.metadata
    import matplotlib.pyplot as plt
    import forsys as fs
    import argparse
    import os
    import numpy as np
    from PIL import Image

    parser = argparse.ArgumentParser(description="Forsys interface command" +
                                     "line interface.")
    parser.add_argument("-f", "--folder", type=str,
                        help="Path to the folder containing the input images.")
    parser.add_argument("-d", "--dynamic", action="store_true",
                        help="Whether to use dynamic inference.\
                        Default is static")
    parser.add_argument("-m", "--method", type=str,
                        help="Method to use for inference. Options are: " +
                        "'nnls', 'lsq' and 'lsq_linear. Default is 'nnls'",
                        default="nnls")
    parser.add_argument("-vn", "--velocity_normalization", type=float,
                        help="Velocity normalization factor \
                        (only used in dynamic mode). Default is 0.1",
                        default=0.1)
    parser.add_argument("-mt", "--max_time", type=int,
                        help="Maximum time step to infer. Default is 0",
                        default=0)
    parser.add_argument("-sf", "--save_folder", type=str,
                        help="Path to the folder in which to save the output. \
                        Default is the same as the input folder.",
                        default=None)
    parser.add_argument("-cc", "--composite", action="store_true",
                        help="Whether to create composite image with the \
                        original microscopies")
    parser.add_argument("-p", "--pngs", action="store_true",
                        help="Whether to create forsys output of each frame")
    parser.add_argument("-c", "--connections", type=str,
                        help="Path to the connections file (optional)",
                        default=None)
    parser.add_argument("-ar", "--aspect_ratio", type=float,
                        help="Aspect ratio of the PNG images. Default is 1",
                        default=1)
    parser.add_argument("-y", "--pressure", action="store_true",
                        help="Plot pressure in the PNG images")
    parser.add_argument("-cb", "--colorbar", action="store_true",
                        help="Plot colorbar in the PNG images")
    parser.add_argument("-v", "--version", action="version",
                        version=f"ForSys version {importlib.metadata.version('forsys')}",
                        help="Show the version of forsys")

    args = parser.parse_args()

    if args.save_folder is not None:
        if not os.path.exists(args.save_folder):
            os.makedirs(args.save_folder)
        if not os.path.exists(os.path.join(args.save_folder, "pngs")):
            os.makedirs(os.path.join(args.save_folder, "pngs"))

        save_to_tiff = os.path.join(args.save_folder,
                                    "forsys_output.tif")
        save_to_png = os.path.join(args.save_folder, "pngs",
                                   "forsys_output.png")
    else:
        save_to_tiff = os.path.join(os.path.dirname(args.folder),
                                    "forsys_output.tif")
        save_to_png = os.path.join(os.path.dirname(args.folder),
                                   "forsys_output.png")
    # DATA_FOLDER = os.path.dirname(args.folder)

    print("This is forsys version ", importlib.metadata.version("forsys"))
    print(f"Running in Dynamic modality {args.dynamic}.",
          f"VN = {args.velocity_normalization}",
          f"Method = {args.method}",)

    print("Inferring...")
    initial_guess = fs.auxiliar.load_initial_guess(args.connections,
                                                   min_time=0,
                                                   max_time=args.max_time)

    microscopies = sorted(list(filter(lambda x: x.endswith(".tif"),
                                      os.listdir(args.folder))))
    if args.dynamic:
        print("Dynamic inference")
        if len(microscopies) == 1:
            print("Only one microscopy found. Inference will be static")

    frames = {}
    time = 0
    image_sizes = []
    original_images = []
    for microscopy in microscopies:
        if args.composite:
            im_ori_fp = Image.open(os.path.join(args.folder, microscopy))
            im_ori = np.array(im_ori_fp)
            original_images.append(im_ori)
            im_ori_fp.close()

        im = Image.open(os.path.join(args.folder, microscopy))
        image_sizes.append(im.size)
        im.close()

        tiff_file = os.path.join(args.folder,
                                 microscopy[:-4],
                                 "handCorrection.tif")

        skeleton = fs.skeleton.Skeleton(tiff_file, mirror_y=False)
        vertices, edges, cells = skeleton.create_lattice()
        vertices, edges, cells, _ = fs.virtual_edges.generate_mesh(vertices,
                                                                   edges,
                                                                   cells,
                                                                   ne=5)
        frames[time] = fs.frames.Frame(time,
                                       vertices,
                                       edges,
                                       cells,
                                       time=time)
        forsys = fs.ForSys(frames, initial_guess=initial_guess,)

        if args.dynamic:
            all_velocities = forsys.get_system_velocity_per_frame()
            ave_velocity_to_normalize = np.mean(all_velocities)
        else:
            ave_velocity_to_normalize = 1

        velocity_normalization = ((args.velocity_normalization /
                                  ave_velocity_to_normalize)
                                  if args.dynamic else 0)
        forsys.build_force_matrix(when=time)
        forsys.solve_stress(when=time,
                            b_matrix="dynamic" if args.dynamic else "static",
                            velocity_normalization=velocity_normalization,
                            adimensional_velocity=False,
                            method=args.method,
                            use_std=False,
                            allow_negatives=False)
        forsys.build_pressure_matrix(when=time)
        forsys.solve_pressure(when=time, method="lagrange_pressure")
        if time == args.max_time:
            break
        time += 1

    print("Plotting...")
    if args.pngs:
        for time in frames.keys():
            fig, ax = plt.subplots(1, 1, figsize=(8, 3))
            fs.plot.plot_inference(forsys.frames[time],
                                   normalized="max",
                                   mirror_y=False,
                                   colorbar=args.colorbar,
                                   pressure=args.pressure,
                                   aspect_ratio=args.aspect_ratio,
                                   ax=ax)
            ax.set_title(f"Time: {time}")
            ax.axis("off")
            plt.tight_layout()

            plt.savefig(save_to_png.replace(".png", f"_{time}.png"), dpi=500)

    # Now tiff
    fs.plot.plot_inference_as_tiff(forsys,
                                   image_sizes=image_sizes,
                                   original_images=original_images,
                                   layers=2,
                                   external_color=0,
                                   save_folder=save_to_tiff)



from dataclasses import dataclass, field
import numpy as np
import json
from typing import Union, Tuple

from  forsys.exceptions import DifferentTissueException

import forsys.vertex as fvertex
import forsys.frames as fframes

### create a mapping between all meshes in a time series
@dataclass
class TimeSeries():
    """Class to generate and access dynamical inference objects. 
    From the given time series' dictionary creates the mesh and link vertices through time.

    :param time_series: Dictionary with all frames to be linked.
    :type time_series: Dictionary with all frames to be linked.
    :param cm: If True, all frames are converted to the CM before linking
    :type cm: bool
    :param initial_guess: Optional dictionary to give linking values before automating. \
        Each key corresponds to a frame, and within has a dictionary with linking edges \
            to their new ids in that frame
    :type initial_guess: dict

    :raises DifferentTissueException: If two tissues are two different to link them, this exception occurs
    """
    time_series: dict
    cm: bool=True
    initial_guess: dict = field(default_factory=dict)

    def __post_init__(self):
        """Constructor method
        """
        self.cutoff = 0.1
        self.accelerations = {}
        self.velocities = {}
        if not self.initial_guess:
            self.initial_guess = {k: {} for k in self.time_series.keys()}


        self.mapping = {}
        for key in self.time_series.keys():
            # print("meshing ", key)
            t0 = self.time_series[key]
            try:
                t1 = self.time_series[key+1]
            except KeyError:
                break
            try:
                self.mapping[key] = self.create_mapping(t0, t1, self.initial_guess[key])
            except DifferentTissueException:
                print("Tissues between ", key, " and ", key+1, " too different, skipping...")
                self.mapping[key] = None


    def export_mapping(self, where: str = 'meshing.txt') -> None:
        """
        Export the mapping built by the class

        :param where: Path to export to, must include filename, defaults to 'meshing.txt'
        :type where: str, optional
        """
        with open(where, 'w') as f:
            json.dump(self.mapping, f)

    @staticmethod
    def get_cm_coords(vertices: dict) -> np.ndarray:
        """Get the coordinate of the center of mass for a dictionary of vertices

        :param vertices: Dictionary with vertics
        :type vertices: dict
        :return: Array with the x and y coordinates of the center of mass, rounded to 3 decimals
        :rtype: np.ndarray
        """
        return np.around([np.mean([v.x for v in vertices.values()]), 
                        np.mean([v.y for v in vertices.values()])], 3)

    def create_mapping(self, t0: fframes.Frame, t1: fframes.Frame, initial_guess: dict) -> dict:
        """
        Create the mapping between two frames by finding the most probable vertex by 
        proximity. If a vertex cannot be mapped, its assigned a None value in the dictionary

        :param t0: Previous frame
        :type t0: fframes.Frame
        :param t1: Current frame
        :type t1: fframes.Frame
        :param initial_guess: Initial guess to link vertices before calculating by proximity
        :type initial_guess: dict
        :raises DifferentTissueException: If two frames are too different, raise exception
        :return: Dictionary with the mapping between the two frames.
        :rtype: dict
        """
        # Mesh should have time two times, t0 and t1
        # If cm = True, move all vertices to the center of mass
        # TODO: remove initial_guess as parameter
        if self.cm:
            self.cm_coords = {}
            self.cm_coords[0] = np.around([np.mean([v.x for v in t0.vertices.values()]), 
                                np.mean([v.y for v in t0.vertices.values()])], 3)
            self.cm_coords[1] = np.around([np.mean([v.x for v in t1.vertices.values()]), 
                                np.mean([v.y for v in t1.vertices.values()])], 3)
            
            for _, v in t0.vertices.items():
                v.x -= self.cm_coords[0][0]
                v.y -= self.cm_coords[0][1]
            for _, v in t1.vertices.items():
                v.x -= self.cm_coords[1][0]
                v.y -= self.cm_coords[1][1]    

        # Only use the real vertices
        realVertices0_ids = np.array([[x[0]]+[x[-1]] for x in t0.big_edges_list])
        realVertices1_ids = np.array([[x[0]]+[x[-1]] for x in t1.big_edges_list])

        externalVertices0_ids = np.array(t0.border_vertices)
        externalVertices1_ids = np.array(t1.border_vertices)

        # take the corresponding dictionary
        rvertices0 = {key: value for key, value in t0.vertices.items() if key in realVertices0_ids.flatten()}
        rvertices1 = {key: value for key, value in t1.vertices.items() if key in realVertices1_ids.flatten()}
        evertices0 = {key: value for key, value in t0.vertices.items() if key in externalVertices0_ids.flatten()}
        evertices1 = {key: value for key, value in t1.vertices.items() if key in externalVertices1_ids.flatten()}

        xcoords0 = [v.x for v in rvertices0.values()]
        ycoords0 = [v.y for v in rvertices0.values()]
        xcoords1 = [v.x for v in rvertices1.values()]
        ycoords1 = [v.y for v in rvertices1.values()]

        if len(xcoords0) > 0:
            minx = min(xcoords0+xcoords1)
            maxx = max(xcoords0+xcoords1)
            miny = min(ycoords0+ycoords1)
            maxy = max(ycoords0+ycoords1)
            self.maxcoord = max(maxx - minx, maxy - miny)

            maxDifference = 0.10
            xshape0 = max(xcoords0) - min(xcoords0)
            yshape0 = max(ycoords0) - min(ycoords0)
            xshape1 = max(xcoords1) - min(xcoords1)
            yshape1 = max(ycoords1) - min(ycoords1)

            displacement = np.sqrt((xshape1 - xshape0)**2 + (yshape1 - yshape0)**2)

            if displacement > maxDifference * self.maxcoord or \
                len(t0.cells) > len(t0.cells) + 1 or \
                len(t1.cells) > len(t1.cells) + 1:           
                raise DifferentTissueException()

        mapping = {}
        # Add the initial guess
        # py3.5+ compatible
        mapping = {**mapping, **initial_guess}

        # find mapping between real vertices
        for v0 in rvertices0.values():
            if v0.id not in mapping.keys():
                try:
                    mapping[v0.id] = self.find_best(v0, rvertices1, mapping.values()).id
                except AttributeError:
                    mapping[v0.id] = None

        # find mapping between externals
        for v0 in evertices0.values():
            if v0.id not in mapping.keys():
                try:
                    mapping[v0.id] = self.find_best(v0, evertices1, mapping.values()).id
                except AttributeError:
                    mapping[v0.id] = None
        return mapping
    
    def find_best(self, v0: fvertex.Vertex, pool: dict, found: list) -> Union[fvertex.Vertex, None]:
        """
        Find the best possible future vertex by proximity, using increasing distances

        :param v0: Vertex at initial time
        :type v0: fvertex.Vertex
        :param pool: Pool of possible vertices at later time
        :type pool: dict
        :param found: Already assigned vertices which are avoided during the search
        :type found: list
        :return: Vertex object of the best possible future position
        :rtype: Union[fvertex.Vertex, None]
        """
        candidatesObverse = []
        candidatesInverse = []
        spread = 0.005
        while len(candidatesObverse) <= 1 and spread < self.cutoff:
            maxspread = spread * self.maxcoord
            for v1 in pool.values():
                # make the map inyective
                if v1.id not in found:
                    xcoord = (v1.x - v0.x)**2
                    ycoord = (v1.y - v0.y)**2
                    if xcoord + ycoord < maxspread**2:
                        candidatesObverse.append(v1)
            spread += spread
        while len(candidatesInverse) <= 1 and spread < self.cutoff:
            for v1 in reversed(pool.values()):
                # make the map inyective
                if v1.id not in found:
                    xcoord = (v1.x - v0.x)**2
                    ycoord = (v1.y - v0.y)**2
                    if xcoord + ycoord < maxspread**2:
                        candidatesInverse.append(v1)
            spread += spread

        candidates = np.concatenate((candidatesInverse, candidatesObverse))
        if len(candidates) == 0:
            best = None
        # from candidates pick the closest
        elif len(candidates) == 1:
            best = candidates[0]
        else:
            distances = [self.distance(v0, vc) for vc in candidates]
            best = candidates[distances.index(min(distances))]
        return best
    
    @staticmethod
    def distance(v0: fvertex.Vertex, v1: fvertex.Vertex) -> float:
        """
        Calculate the distance between two vertices

        :param v0: First vertex object
        :type v0: fvertex.Vertex
        :param v1: Second vertex object
        :type v1: fvertex.Vertex
        :return: Euclidean distance between vertices
        :rtype: float
        """
        return np.sqrt((v1.x - v0.x)**2 + (v1.y - v0.y)**2)

    def calculate_velocity(self, point: int, initial_time: str) -> np.ndarray:
        """
        Calculate the velocity of a given vertex. Forward finite difference is used, 
        except the last point where backward is used.

        :param point: ID of the desired vertex at the initial time
        :type point: int
        :param initial_time: Frame number to use as initial
        :type initial_time: str
        :raises DifferentTissueException: _description_
        :raises DifferentTissueException: _description_
        :return: X and Y components of the velocity
        :rtype: np.ndarray
        """
        v0 = self.time_series[initial_time].vertices[point]
        if initial_time == len(self.time_series) - 1:
            # last time, use backward
            if self.mapping[initial_time - 1] == None:
                raise DifferentTissueException()
            tt1 = initial_time - 1
        else:
            tt1 = initial_time + 1

            if self.mapping[initial_time] == None:
                raise DifferentTissueException()


        try:
            v1 = self.time_series[tt1].vertices[self.get_point_id_by_map(point, initial_time, tt1)]
        except KeyError:
            # fictious vertex to give velocity zero
            v1 = fvertex.Vertex(-1, v0.x, v0.y)
        

        try:
            ti = self.time_series[int(initial_time)].time
            tf = self.time_series[tt1].time
        except KeyError:
            tf = 1
            ti = 0

        return (np.array([v1.x, v1.y]) - np.array([v0.x, v0.y])) / (tf - ti)

    def calculate_displacement(self, point: int, initial_time: str) -> np.ndarray:
        """
        Calculate the velocity of a given vertex. Forward finite difference is used, 
        except the last point where backward is used.

        :param point: ID of the desired vertex at the initial time
        :type point: int
        :param initial_time: Frame number to use as initial
        :type initial_time: str
        :raises DifferentTissueException: _description_
        :raises DifferentTissueException: _description_
        :return: X and Y components of the velocity
        :rtype: np.ndarray
        """
        v0 = self.time_series[initial_time].vertices[point]
        if initial_time == len(self.time_series) - 1:
            # last time, use backward
            if self.mapping[initial_time - 1] == None:
                raise DifferentTissueException()
            tt1 = initial_time - 1
        else:
            tt1 = initial_time + 1

            if self.mapping[initial_time] == None:
                raise DifferentTissueException()


        try:
            v1 = self.time_series[tt1].vertices[self.get_point_id_by_map(point, initial_time, tt1)]
        except KeyError:
            # fictious vertex to give velocity zero
            v1 = fvertex.Vertex(-1, v0.x, v0.y)
        

        try:
            ti = self.time_series[int(initial_time)].time
            tf = self.time_series[tt1].time
        except KeyError:
            tf = 1
            ti = 0

        return (np.array([v1.x, v1.y]) - np.array([v0.x, v0.y]))

    def whole_tissue_acceleration(self, initial_time: int) -> dict:
        """
        Calculate an estimation of the acceleration in the whole tissue, per big edge

        :param initial_time: Initial time
        :type initial_time: int
        :return: Acceleration of all big edges
        :rtype: dict
        """
        t0 = self.time_series[initial_time]
        for ev in t0.big_edges_list:
            acc0 = self.calculate_acceleration(ev[0], initial_time)
            acc1 = self.calculate_acceleration(ev[-1], initial_time)
            self.accelerations[t0.big_edges_list.index(ev)] = np.linalg.norm(acc0) + np.linalg.norm(acc1)
        return self.accelerations

    def whole_tissue_velocity(self, initial_time: int) -> dict:
        """
        Calculate an estimation of the velocity in the whole tissue, per big edge

        :param initial_time: Initial time
        :type initial_time: int
        :return: Dictionary with the velocity per big edge
        :rtype: dict
        """
        t0 = self.time_series[initial_time]
        for ev in t0.big_edges_list:
            vel0 = self.calculate_velocity(ev[0], initial_time)
            vel1 = self.calculate_velocity(ev[-1], initial_time)
            self.velocities[t0.big_edges_list.index(ev)] = np.linalg.norm(vel0) + np.linalg.norm(vel1)
        return self.velocities


    def calculate_acceleration(self, point: int, initial_time: int) -> list:
        """
        Calculate acceleration of a vertex in time. At least three timepoints are required
        for calcualting the acceleration. Forward finite difference is used, except for the 
        last points where either backward or central is used.

        :param point: Vertex to use
        :type point: int
        :param initial_time: Initial time
        :type initial_time: int
        :return: Acceleration of the given vertex
        :rtype: list
        """
        # This requires three timepoints
        # TODO What if the timepoints are not equally distributed ? (if using finite differences)
        t0 = self.time_series[initial_time]        
        if initial_time == len(self.time_series) - 1:
            # last time, use backward
            tt1 = initial_time - 1
            tt2 = initial_time - 2            
        elif initial_time == 0:
            # first time, use forward
            tt1 = initial_time + 1
            tt2 = initial_time + 2
        else:
            # Middle, use central
            tt1 = initial_time + 1
            tt2 = initial_time - 1
        
        t1 = self.time_series[tt1]
        t2 = self.time_series[tt2]
        v0 = t0.vertices[point]

        try:
            v1 = t1.vertices[self.get_point_id_by_map(point, initial_time, tt1)]
            v2 = t2.vertices[self.get_point_id_by_map(point, initial_time, tt2)]
            
            f0 = np.array([v0.x, v0.y])
            f1 = np.array([v1.x, v1.y])
            f2 = np.array([v2.x, v2.y])

            # TODO: Generalize to non consecutive times ?
            if initial_time == len(self.time_series) - 1:
                acceleration = f0 - 2 * f1 + f2
            elif initial_time == 0:
                acceleration = f2 - 2 * f1 + f0
            else:
                acceleration = f1 - 2 * f0 + f2


        except KeyError:
            # that vertex has no counterpart at a later time
            acceleration = np.nan

        try:
            ti = t0.time
            tf = t2.time
        except KeyError:
            tf = 1
            ti = 0
        return acceleration

    def get_point_id_by_map(self, point: int, initial_time: int, final_time: int) -> int:
        """
        Get the ID of a vertex at any later time, given one initial.

        :param point: Point ID at initial time
        :type point: int
        :param initial_time: Initial time
        :type initial_time: int
        :param final_time: Final time to get the ID
        :type final_time: int
        :return: ID of the required vertex at time final_time
        :rtype: int
        """
        if initial_time < final_time:
            timerange = np.arange(initial_time, final_time, 1)
        else:
            timerange = np.arange(final_time, initial_time, 1)[::-1]
        
        for ii in timerange:
            if initial_time < final_time:
                tempMapping = self.mapping[ii]
            else:
                tempMapping = {v: k for k, v in self.mapping[ii].items()}
            if point == None or tempMapping == None:
                break
            point = tempMapping[point]            
        
        return point
        
    def acceleration_per_edge(self, earrid: int, initial_time: int, final_time: int) -> list:
        """
        Get list with edge acceleration from an initial time to a final time. 
        Edge accelration is the sum of the pivot vertices acceleration.

        :param earrid: ID of the big edge of interest
        :type earrid: int
        :param initial_time: Initial time
        :type initial_time: int
        :param final_time: Final time
        :type final_time: int
        :return: Row with the acceleration values
        :rtype: list
        """
        t0 = self.time_series[initial_time]

        earr = t0.big_edges_list[earrid]
        timerange = np.arange(initial_time, final_time, 1)
        row = []
        for ii in timerange:
            id0 = self.get_point_id_by_map(earr[0], initial_time, ii)
            id1 = self.get_point_id_by_map(earr[-1], initial_time, ii)
            if id0 == None or id1 == None:
                acc = np.nan
            else:
                try:
                    acc0 = self.calculate_acceleration(id0, ii)
                    acc1 = self.calculate_acceleration(id1, ii)
                    acc = np.linalg.norm(acc0) + np.linalg.norm(acc1)
                except DifferentTissueException:
                    acc = np.nan

            row.append(acc)
        return row

    def velocity_per_edge(self, earrid: int, initial_time: int, final_time: int) -> list:
        """
        Get list with edge velocity from an initial time to a final time. 
        Edge velocity is the sum of the pivot vertices velocities.

        :param earrid: ID of the big edge of interest
        :type earrid: int
        :param initial_time: Initial time
        :type initial_time: int
        :param final_time: Final time
        :type final_time: int
        :return: Row with the acceleration values
        :rtype: list
        """
        #Return the acceleration of the earrid in the given timeframe
        t0 = self.time_series[initial_time]

        earr = t0.big_edges_list[earrid]
        timerange = np.arange(initial_time, final_time, 1)
        row = []
        for ii in timerange:
            id0 = self.get_point_id_by_map(earr[0], initial_time, ii)
            id1 = self.get_point_id_by_map(earr[-1], initial_time, ii)
            if id0 == None or id1 == None:
                vel = np.nan
            else:
                try:
                    vel0 = self.calculate_velocity(id0, ii)
                    vel1 = self.calculate_velocity(id1, ii)
                    vel = np.linalg.norm(vel0) + np.linalg.norm(vel1)
                except DifferentTissueException:
                    vel = np.nan

            row.append(vel)
        return row

    def times_to_use(self, last_frame: bool = False) -> list:
        """
        Get list of times to use

        :param last_frame: Include up until the last frame, defaults to False
        :type last_frame: bool, optional
        :return: List of times
        :rtype: list
        """

        final_times = [-1]
        if not last_frame:
            t_range = range(0, len(self.time_series) - 1)
        else:
            t_range = range(0, last_frame-1)
            
        for t in t_range:
            if self.mapping[t] == None:
                final_times.append(t)
        final_times.append(t+1)
        return final_times

    def get_vertex_position(self, v0: int, t0: int, tmax: int) -> Tuple:
        """
        Get the position of a vertex for all times from t0 to tmax.

        :param v0: Vertex ID
        :type v0: int
        :param t0: Initial time
        :type t0: int
        :param tmax: Final time
        :type tmax: int
        :return: Tuple with the X and Y position of the vertex
        :rtype: Tuple
        """
        positions_x = []
        positions_y = []
        if tmax == -1:
            range_values = self.time_series.keys()
        else:
            range_values = range(t0, tmax)
        for t in range_values:
            print(f"Vertex {v0} is now {self.get_point_id_by_map(v0, t0, t)}")
            v_now = self.time_series[t].vertices[self.get_point_id_by_map(v0, t0, t)]
            positions_x.append(v_now.x)
            positions_y.append(v_now.y)

        return positions_x, positions_y
